"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import inspect
import json
import time
from pathlib import Path
from typing import Generator, Optional, List, Iterable

import botocore
from botocore.exceptions import ClientError
from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.ec2_connection')


def create_permissions(ranges):
    ip_ranges = [{'CidrIp': ip_range} for ip_range in ranges]
    return [{
        'IpProtocol': '-1',
        'FromPort': -1,
        'ToPort': -1,
        'IpRanges': ip_ranges
    }]


def preserve_default_permission(group_id, permissions):
    for permission in permissions:
        if 'UserIdGroupPairs' in permission and permission['UserIdGroupPairs']:
            for pair in permission['UserIdGroupPairs']:
                if pair['GroupId'] == group_id:
                    permission.pop('UserIdGroupPairs')


@apply_methods_decorator(retry())
class EC2Connection(object):
    """ EC2 connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('ec2', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new EC2 connection.')

    def describe_security_groups(self, name=None, sg_id=None, vpc_id=None):
        filters = []
        if name:
            if isinstance(name, list):
                filters.append({'Name': 'group-name', 'Values': name})
            elif isinstance(name, str):
                filters.append({'Name': 'group-name', 'Values': [name]})
            else:
                _LOG.warn('Unacceptable name type: %s', type(name))
        if sg_id:
            if isinstance(sg_id, list):
                filters.append({'Name': 'group-id', 'Values': sg_id})
            elif isinstance(sg_id, str):
                filters.append({'Name': 'group-id', 'Values': [sg_id]})
            else:
                _LOG.warn(
                    f'Unacceptable security group id type: {type(sg_id)}')
        if vpc_id:
            filters.append({'Name': 'vpc-id', 'Values': [vpc_id]})
        return self.client.describe_security_groups(Filters=filters)[
            'SecurityGroups']

    def describe_regions(self, name=None):
        filters = []
        if name:
            if isinstance(name, list):
                filters.append({'Name': 'region-name', 'Values': name})
            elif isinstance(name, str):
                filters.append({'Name': 'region-name', 'Values': [name]})
            else:
                _LOG.warn('Unacceptable name type: %s', type(name))
        return self.client.describe_regions(Filters=filters)['Regions']

    def get_default_vpc_id(self):
        for vpc in self.client.describe_vpcs()['Vpcs']:
            if vpc['IsDefault']:
                return vpc['VpcId']

    def create_sg(self, name, desc, vpc_id):
        return self.client.create_security_group(GroupName=name,
                                                 Description=desc,
                                                 VpcId=vpc_id)['GroupId']

    def authorize_ingress(self, group_id, group_name, ranges):
        if not group_id:
            group_id = self.get_sg_id(group_name)
        permissions = create_permissions(ranges)
        self.client.authorize_security_group_ingress(GroupId=group_id,
                                                     GroupName=group_name,
                                                     IpPermissions=permissions)

    def revoke_ingress(self, group_id, group_name, ranges):
        permissions = create_permissions(ranges)
        self.revoke_ingress_permissions(group_id, group_name, permissions)

    def revoke_ingress_permissions(self, group_id, group_name, permissions):
        if not group_id:
            group_id = self.get_sg_id(group_name)
        preserve_default_permission(group_id, permissions)
        if not permissions:
            return
        self.client.revoke_security_group_ingress(GroupId=group_id,
                                                  GroupName=group_name,
                                                  IpPermissions=permissions)

    def delete_sg(self, group_name):
        self.client.delete_security_group(GroupName=group_name)

    def get_sg_id(self, group_name, vpc_id=None):
        if not vpc_id:
            vpc_id = self.get_default_vpc_id()
        group, = self.describe_security_groups(group_name, vpc_id)
        if group:
            return group['GroupId']

    def get_key_pairs(self, dry_run=False, key_names=None, filters=None):
        """
        :type dry_run: bool
        :type key_names: list
        :type filters: list
        :return:
        """
        params = dict(DryRun=dry_run)
        if key_names:
            params['KeyNames'] = key_names
        if filters:
            params['Filters'] = filters
        return self.client.describe_key_pairs(**params)

    def if_key_pair_exists(self, key_name):
        key_pairs = self.get_key_pairs().get('KeyPairs')
        if key_pairs:
            for each in key_pairs:
                if each['KeyName'] == key_name:
                    return True

    def list_vpcs(self, dry_run=None, vpc_ids=None, filters=None):
        params = dict()
        if dry_run:
            params['DryRun'] = dry_run
        if vpc_ids:
            params['VpcIds'] = vpc_ids
        if filters:
            params['Filters'] = filters
        response = self.client.describe_vpcs(**params)
        if response:
            return response['Vpcs']

    def list_subnets(self, dry_run=None, subnet_ids=None, filters=None):
        params = dict()
        if dry_run:
            params['DryRun'] = dry_run
        if subnet_ids:
            params['SubnetIds'] = subnet_ids
        if filters:
            params['Filters'] = filters
        response = self.client.describe_subnets(**params)
        if response:
            return response['Subnets']

    def get_azs(self):
        response = self.client.describe_availability_zones()
        return [az['ZoneName'] for az in response['AvailabilityZones']]

    def describe_image(self, image_id):
        params = dict(ImageIds=[image_id])
        response = self.client.describe_images(**params)
        return response['Images']

    def describe_instances(self, filters):
        result_list = []
        response = self.client.describe_instances(Filters=filters)
        result_list.extend([reservation['Instances'][0]
                            for reservation in response['Reservations']])
        token = response.get('NextToken')
        while token:  # value is 'null' if there is no token
            response = self.client.describe_instances(Filters=filters,
                                                      NextToken=token)
            result_list.extend([reservation['Instances'][0]
                                for reservation in response['Reservations']])
            token = response.get('NextToken')
        return result_list

    def terminate_instances(self, instance_ids):
        self.client.terminate_instances(
            InstanceIds=instance_ids
        )

    def launch_instance(self, image_id, instance_type,
                        security_groups_names=None,
                        security_group_ids=None,
                        iam_instance_profile=None,
                        name=None, key_name=None,
                        user_data=None, tags_list=None,
                        subnet_id=None, availability_zone=None):

        if iam_instance_profile:
            if not iam_instance_profile.get('Arn') \
                    and not iam_instance_profile.get('Name'):
                raise AssertionError('Provided instance profile {0}'
                                     'is not well-formed. '
                                     'Arn or Name nodes required.'
                                     .format(iam_instance_profile))

        tag_specification = {
            'ResourceType': 'instance',
        }
        name_tag = None
        if name:
            name_tag = {
                'Key': 'Name',
                'Value': name
            }

        if tags_list:
            valid_tags = [tag for tag in tags_list
                          if tag['Key'] and tag['Value']]
            if not valid_tags:
                _LOG.warn('There are no valid tags in tags_list specified')
            else:
                if name_tag:
                    valid_tags.append(name_tag)
                tag_specification['Tags'] = valid_tags
        else:
            tag_specification['Tags'] = [name_tag]

        instance_parameters = {
            'ImageId': image_id,
            'InstanceType': instance_type,
            'TagSpecifications': [tag_specification],
            'MinCount': 1,
            'MaxCount': 1
        }

        if availability_zone:
            instance_parameters['Placement'] = {
                'AvailabilityZone': availability_zone
            }
        if key_name:
            instance_parameters['KeyName'] = key_name
        if user_data:
            instance_parameters['UserData'] = user_data
        if iam_instance_profile:
            instance_parameters['IamInstanceProfile'] = iam_instance_profile
        if subnet_id:
            instance_parameters['SubnetId'] = subnet_id
        if security_groups_names:
            instance_parameters['SecurityGroups'] = security_groups_names
        if security_group_ids:
            instance_parameters['SecurityGroupIds'] = security_group_ids

        response = self.client.run_instances(**instance_parameters)
        # always launch only one instance
        launched_instances = response['Instances']
        if len(launched_instances) < 1:
            return 'No instances launched'
        else:
            return launched_instances[0]

    def modify_instance_attribute(self, **kwargs):
        if not kwargs['InstanceId']:
            raise AssertionError('InstanceId must be specified')
        self.client.modify_instance_attribute(**kwargs)

    def deploy_security_groups(self, groups):
        default_vpc_id = self.get_default_vpc_id()
        groups_in_default_vpc = self.describe_security_groups(
            vpc_id=default_vpc_id)
        group_names = [g['GroupName'] for g in groups_in_default_vpc]

        verify = []
        create = []
        for group in groups:
            if group['n'] in group_names:
                verify.append(group)
            else:
                create.append(group)

        if create:
            self._create_security_groups(default_vpc_id, create)

        if verify:
            self._verify_security_groups(verify, groups_in_default_vpc)

    def _delete_security_groups(self, groups):
        vpc_id = self.client.get_default_vpc_id()
        security_groups = self.client.describe_security_groups(groups, vpc_id)
        existent_names = [sg['GroupName'] for sg in security_groups]
        for sg_name in groups:
            if sg_name not in existent_names:
                continue
            if sg_name == 'default':
                security_group, = self.client.describe_security_groups(
                    'default', vpc_id)
                self.client.revoke_ingress_permissions(group_id=None,
                                                       group_name='default',
                                                       permissions=
                                                       security_group[
                                                           'IpPermissions'])
            else:
                self.client.delete_sg(sg_name)

    def _create_security_groups(self, vpc_id, groups_to_create):
        for group in groups_to_create:
            if group['n'] == 'default':
                continue
            group_id = self.create_sg(group['n'], group['d'], vpc_id)
            ranges = group['r']
            # add waiting to auth sg
            time.sleep(5)
            self.authorize_ingress(group_id, group['n'], ranges)

    def _verify_security_groups(self, groups, sgs):
        security_groups = {sg['GroupName']: sg for sg in sgs}

        for group in groups:
            name = group['n']
            sg = security_groups[name]
            should_be = group['r']
            actual = []

            for permission in sg['IpPermissions']:
                for ip_range in permission['IpRanges']:
                    actual.append(ip_range['CidrIp'])

            to_revoke = [cidr for cidr in actual if cidr not in should_be]
            to_authorize = [cidr for cidr in should_be if cidr not in actual]

            if to_revoke:
                self.revoke_ingress(sg['GroupId'], name, to_revoke)
            if to_authorize:
                # add waiting to auth sg
                time.sleep(5)
                self.authorize_ingress(sg['GroupId'], name,
                                       to_authorize)

    def associate_address(self, instance_id=None, public_ip=None,
                          allow_reassociation=False):
        params = dict(AllowReassociation=allow_reassociation)
        if instance_id:
            params['InstanceId'] = instance_id
        if public_ip:
            params['PublicIp'] = public_ip
        return self.client.associate_address(**params)

    def create_launch_template(self, name, lt_data, version_description=None,
                               tag_specifications=None):
        params = dict()
        params['LaunchTemplateName'] = name
        params['LaunchTemplateData'] = lt_data
        if version_description is not None:
            params['VersionDescription'] = version_description
        if tag_specifications is not None:
            params['TagSpecifications'] = tag_specifications
        return self.client.create_launch_template(**params)

    def create_launch_template_version(self, lt_name=None, lt_id=None,
                                       source_version=None, lt_data=None,
                                       version_description=None):
        params = dict()
        if source_version is not None:
            params['SourceVersion'] = source_version
        if lt_data is not None:
            params['LaunchTemplateData'] = lt_data
        if version_description is not None:
            params['VersionDescription'] = version_description

        if lt_name is not None and lt_id is not None:
            _LOG.warn('Both the launch template name and ID are specified. '
                      'The request will be made by ID.')
            params['LaunchTemplateId'] = lt_id

        elif lt_name is not None:
            params['LaunchTemplateName'] = lt_name

        elif lt_id is not None:
            params['LaunchTemplateId'] = lt_id

        else:
            _LOG.error('A launch template version can not be created without '
                       'the name or ID of the launch template.')
            return
        return self.client.create_launch_template_version(**params)

    def describe_launch_templates(self, lt_name=None, lt_id=None):
        result_list = list()
        params = dict()
        if lt_name is not None and lt_id is not None:
            _LOG.warn('Both the launch template name and ID are specified. '
                      'The request will be made by ID.')
            if isinstance(lt_id, list):
                params['LaunchTemplateIds'] = lt_id
            elif isinstance(lt_id, str):
                params['LaunchTemplateIds'] = [lt_id]
            else:
                _LOG.warn(
                    f'Unsupported launch template ID type {type(lt_id)}')

        elif lt_name is not None:
            if isinstance(lt_name, list):
                params['LaunchTemplateNames'] = lt_name
            elif isinstance(lt_name, str):
                params['LaunchTemplateNames'] = [lt_name]
            else:
                _LOG.warn(
                    f'Unsupported launch template name type {type(lt_name)}')

        elif lt_id is not None:
            if isinstance(lt_id, list):
                params['LaunchTemplateIds'] = lt_id
            elif isinstance(lt_id, str):
                params['LaunchTemplateIds'] = [lt_id]
            else:
                _LOG.warn(
                    f'Unsupported launch template ID type {type(lt_id)}')
        try:
            response = self.client.describe_launch_templates(**params) if (
                params) else self.client.describe_launch_templates()
            token = response.get('NextToken')
            result_list.extend(response['LaunchTemplates'])
            while token:
                if params:
                    params['NextToken'] = token
                    response = self.client.describe_launch_templates(**params)
                else:
                    response = self.client.describe_launch_templates()
                token = response.get('NextToken')
                result_list.extend(response['LaunchTemplates'])
        except ClientError as e:
            if ('InvalidLaunchTemplateName.NotFoundException' in str(e) or
                    'InvalidLaunchTemplateId.NotFound' in str(e)):
                dynamic_message = f"by name '{lt_name}'" if lt_name else \
                    f"by ID '{lt_id}'"
                _LOG.warn(f"Launch template not found by {dynamic_message}")
            else:
                raise e
        return result_list

    def delete_launch_template(self, lt_name=None, lt_id=None):
        params = dict()
        if lt_name is not None and lt_id is not None:
            _LOG.warn('Both the launch template name and ID are specified. '
                      'The request will be made by ID.')
            params['LaunchTemplateId'] = lt_id

        elif lt_name is not None:
            params['LaunchTemplateName'] = lt_name

        elif lt_id is not None:
            params['LaunchTemplateId'] = lt_id
        else:
            raise AssertionError(
                'Either the launch template name or ID must be provided for '
                'removing the launch template.')
        self.client.delete_launch_template(**params)

    def modify_launch_template(self, default_version, lt_name=None,
                               lt_id=None):
        params = dict()
        params['DefaultVersion'] = default_version

        if lt_name is not None and lt_id is not None:
            _LOG.warn('Both the launch template name and ID are specified. '
                      'The request will be made by ID.')
            params['LaunchTemplateId'] = lt_id

        elif lt_name is not None:
            params['LaunchTemplateName'] = lt_name

        elif lt_id is not None:
            params['LaunchTemplateId'] = lt_id

        else:
            _LOG.error('A launch template modification can not be done '
                       'without the name or ID of the launch template.')
            return
        return self.client.modify_launch_template(**params)


class InstanceTypes:
    @staticmethod
    def from_api(region_name: Optional[str] = None,
                 current_generation: Optional[bool] = None,
                 arch: Optional[List] = None,
                 ) -> Generator[str, None, None]:
        filters = []
        if isinstance(current_generation, bool):
            filters.append({
                'Name': 'current-generation',
                'Values': [str(current_generation).lower()]
            })
        if isinstance(arch, list):
            assert set(arch).issubset({'arm64', 'i386', 'x86_64'})
            filters.append({
                'Name': 'processor-info.supported-architecture',
                'Values': arch
            })

        params = {}
        if filters:
            params['Filters'] = filters

        ec2 = client('ec2', region_name=region_name)
        while True:
            res = ec2.describe_instance_types(**params)
            yield from (item['InstanceType'] for item in res['InstanceTypes'])
            _next = res.get('NextToken')
            if not _next:
                break
            params['NextToken'] = _next

    @staticmethod
    def from_botocore() -> Generator[str, None, None]:
        path = Path(inspect.getfile(botocore)).parent
        with open(Path(path, 'data', 'ec2', '2016-11-15',
                       'service-2.json'), encoding="utf8") as fp:
            data = json.load(fp)
        yield from data['shapes']['InstanceType']['enum']

    @staticmethod
    def instance_type_group(instance_type: str) -> str:
        return instance_type.split('.')[0]

    @staticmethod
    def with_groups(it: Iterable[str]) -> Generator[str, None, None]:
        """
        Before yielding an instance type, yields its group.
        A group is yielded only once.
        :param it: Iterable[str]
        :return: Generator[str, None, None]
        """
        emitted = set()
        for instance_type in it:
            group = InstanceTypes.instance_type_group(instance_type)
            if group not in emitted:
                yield group
                emitted.add(group)
            yield instance_type



from pathlib import Path

from syndicate.core.generators import _mkdir, _write_content_to_file
from syndicate.core.generators.contents import PYTHON_TESTS_INIT_CONTENT, \
    PYTHON_TESTS_INIT_LAMBDA_TEMPLATE, PYTHON_TESTS_BASIC_TEST_CASE_TEMPLATE
from syndicate.core.helper import string_to_capitalized_camel_case

PYTHON_TESTS_DIR = 'tests'
PYTHON_BASIC_TEST_CASE_FILE = 'test_success.py'


class PythonTestsGenerator:
    def __init__(self, project_path: str, lambda_name: str):
        self.project_path = project_path
        self.lambda_name = lambda_name
        self.processed_name = self.lambda_name.replace("-", "_")
        self.test_lambda_folder = f'test_{self.processed_name.lower()}'
        self.camel_lambda_name = string_to_capitalized_camel_case(
            self.processed_name)

    def generate(self):
        folder_path = Path(self.project_path, PYTHON_TESTS_DIR)
        init_path = Path(folder_path, '__init__.py')
        lambda_folder_path = Path(
            folder_path, self.test_lambda_folder)
        lambda_init_path = Path(lambda_folder_path, '__init__.py')
        lambda_basic_test_case_path = Path(lambda_folder_path,
                                           PYTHON_BASIC_TEST_CASE_FILE)

        _mkdir(folder_path, exist_ok=True)
        if not init_path.exists():
            _write_content_to_file(init_path, PYTHON_TESTS_INIT_CONTENT)
        _mkdir(lambda_folder_path, exist_ok=True)

        _write_content_to_file(
            lambda_init_path,
            PYTHON_TESTS_INIT_LAMBDA_TEMPLATE.format(
                lambda_name=self.lambda_name,
                camel_lambda_name=self.camel_lambda_name))

        _write_content_to_file(
            lambda_basic_test_case_path,
            PYTHON_TESTS_BASIC_TEST_CASE_TEMPLATE.format(
                test_lambda_folder=self.test_lambda_folder,
                camel_lambda_name=self.camel_lambda_name))


def _generate_python_tests(project_path, lambda_name):
    PythonTestsGenerator(project_path, lambda_name).generate()



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import sqs, GetAtt

from syndicate.core.resources.sqs_resource import SqsResource
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import to_logic_name, sqs_queue_logic_name


class CfSqsConverter(CfResourceConverter):

    def convert(self, name, meta):
        is_fifo = bool(meta.get('fifo_queue', False))
        queue_name = SqsResource.build_resource_name(is_fifo, name)

        delay_sec = meta.get('delay_seconds')
        max_mes_size = meta.get('maximum_message_size')
        mes_ret_period = meta.get('message_retention_period')
        policy = meta.get('policy')
        receive_mes_wait_sec = meta.get('receive_message_wait_time_seconds')
        redrive_policy = meta.get('redrive_policy')
        vis_timeout = meta.get('visibility_timeout')
        kms_master_key_id = meta.get('kms_master_key_id')
        kms_data_reuse_period = meta.get('kms_data_key_reuse_period_seconds')
        content_deduplication = bool(meta.get('content_based_deduplication',
                                              False))

        queue = sqs.Queue(sqs_queue_logic_name(name))
        queue.QueueName = queue_name
        self.template.add_resource(queue)

        if delay_sec:
            if delay_sec < 0 or delay_sec > 900:
                raise AssertionError(
                    'Delay seconds for queue must be '
                    'between 0 and 900 seconds')
            queue.DelaySeconds = delay_sec

        if max_mes_size:
            if max_mes_size < 1024 or max_mes_size > 262144:
                raise AssertionError(
                    'Maximum message size must be '
                    'between 1024 and 262144 bytes')
            queue.MaximumMessageSize = max_mes_size

        if mes_ret_period:
            if mes_ret_period < 60 or mes_ret_period > 1209600:
                raise AssertionError(
                    'Message retention size must be '
                    'between 60 and 1209600 seconds')
            queue.MessageRetentionPeriod = mes_ret_period

        if receive_mes_wait_sec:
            if receive_mes_wait_sec < 0 or receive_mes_wait_sec > 20:
                raise AssertionError(
                    'Receive message wait time must be '
                    'between 0 and 20 seconds')
            queue.ReceiveMessageWaitTimeSeconds = receive_mes_wait_sec

        if redrive_policy:
            queue.RedrivePolicy = sqs.RedrivePolicy(
                deadLetterTargetArn=redrive_policy['deadLetterTargetArn'],
                maxReceiveCount=redrive_policy['maxReceiveCount']
            )

        if vis_timeout:
            if vis_timeout < 0 or vis_timeout > 43200:
                raise AssertionError(
                    'Visibility timeout must be '
                    'between 0 and 43200 seconds')
            queue.VisibilityTimeout = vis_timeout

        if kms_master_key_id:
            queue.KmsMasterKeyId = kms_master_key_id

        if kms_data_reuse_period:
            if kms_data_reuse_period < 60 or kms_data_reuse_period > 86400:
                raise AssertionError(
                    'KMS key reuse period must be '
                    'between 60 and 86400 seconds')
            queue.KmsDataKeyReusePeriodSeconds = kms_data_reuse_period

        if is_fifo:
            queue.FifoQueue = True

        if content_deduplication:
            queue.ContentBasedDeduplication = True

        if policy:
            queue_policy = sqs.QueuePolicy(
                to_logic_name('SQSQueuePolicy', name))
            queue_policy.PolicyDocument = policy
            queue_policy.Queues = [queue.ref()]
            self.template.add_resource(queue_policy)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import uuid

from troposphere import events

from syndicate.core.resources.cloud_watch_resource import \
    validate_cloud_watch_rule_params, get_event_bus_arn
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import cloudwatch_rule_logic_name


def _create_schedule_rule(rule_meta, rule_res):
    expression = rule_meta['expression']
    rule_res.ScheduleExpression = expression
    rule_res.Description = rule_res.Name
    rule_res.State = 'ENABLED'


def _create_ec2_rule(rule_meta, rule_res):
    instances = rule_meta.get('instance_ids'),
    instance_states = rule_meta.get('instance_states')
    event_pattern = {
        "source": ["aws.ec2"],
        "detail-type": ["EC2 Instance State-change Notification"]
    }
    if instances:
        event_pattern["detail"] = {"instance-id": instances}
    if instance_states:
        if event_pattern.get("detail"):
            event_pattern.get("detail").update({"state": instance_states})
        else:
            event_pattern["detail"] = {"state": instance_states}

    rule_res.EventPattern = event_pattern
    rule_res.Description = rule_res.Name
    rule_res.State = 'ENABLED'


def _create_api_call_rule(rule_meta, rule_res):
    aws_service = rule_meta.get('aws_service')
    operations = rule_meta.get('operations')
    custom_pattern = rule_meta.get('custom_pattern')
    if custom_pattern:
        event_pattern = custom_pattern
    elif aws_service:
        event_pattern = {
            'detail-type': [
                'AWS API Call via CloudTrail'
            ],
            'detail': {
                'eventSource': [
                    '{0}.amazonaws.com'.format(aws_service)
                ]
            }
        }
        if operations:
            event_pattern['detail']['eventName'] = operations
    else:
        raise AssertionError(
            'aws_service or custom_pattern should be specified for rule '
            'with "api_call" type! Resource: {0}'.format(rule_res.Name))
    rule_res.EventPattern = event_pattern
    rule_res.Description = rule_res.Name
    rule_res.State = 'ENABLED'


RULE_TYPES = {
    'schedule': _create_schedule_rule,
    'ec2': _create_ec2_rule,
    'api_call': _create_api_call_rule
}


def attach_rule_target(rule, target_arn, target_id=None):
    if target_id is None:
        target_id = str(uuid.uuid1())
    try:
        targets = rule.Targets
    except AttributeError:
        targets = []
        rule.Targets = targets
    targets.append(events.Target(
        Arn=target_arn,
        Id=target_id)
    )


class CfCloudWatchRuleConverter(CfResourceConverter):

    def convert(self, name, meta):
        validate_cloud_watch_rule_params(name=name, meta=meta)

        rule = events.Rule(cloudwatch_rule_logic_name(name))
        rule.Name = name
        self.template.add_resource(rule)

        rule_type = meta['rule_type']
        func = RULE_TYPES[rule_type]
        func(meta, rule)

        event_buses = meta.get('event_bus_accounts')
        if event_buses:
            self._attach_tenant_rule_targets(rule_res=rule,
                                             event_buses=event_buses)

    def _attach_tenant_rule_targets(self, rule_res, event_buses):
        region = self.config.region
        for event_bus in event_buses:
            target_arn = get_event_bus_arn(event_bus=event_bus,
                                           region=region)
            attach_rule_target(rule=rule_res,
                               target_arn=target_arn)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.constants import IAM_AUTH_TYPE, NONE_AUTH_TYPE, \
    LAMBDA_ARCHITECTURE_LIST


class LambdaValidator:
    def __init__(self, name, meta):
        self._name = name,
        self._meta = meta

    def validate(self):
        url_config = self._meta.get('url_config')
        if url_config:
            self._validate_url_config(url_config)
        ephemeral_storage = self._meta.get('ephemeral_storage')
        if ephemeral_storage:
            self._validate_ephemeral_storage(ephemeral_storage)
        architectures = self._meta.get('architectures')
        if architectures:
            self._validate_architecture(architectures)

    def _error(self, message):
        raise AssertionError(message)

    def _validate_url_config(self, url_config):
        if 'auth_type' not in url_config:
            self._error('\'auth_type\' is required in lambda\'s url config')
        auth_types = [IAM_AUTH_TYPE, NONE_AUTH_TYPE]
        if url_config['auth_type'] not in auth_types:
            self._error('\'auth_type\' must be equal to one of these: '
                        f'{", ".join(auth_types)}')
        cors = url_config.get('cors')
        if cors and not isinstance(cors, dict):
            self._error('\'cors\' parameter in lambda config must be a dict')
        if cors:
            allowed_parameters = {'allow_credentials',
                                  'allow_headers', 'allow_methods',
                                  'allow_origins', 'expose_headers',
                                  'max_age'}
            impostors = set(cors.keys()) - allowed_parameters
            if impostors:
                self._error(f'Only these parameters are allowed: '
                            f'{allowed_parameters}')

    def _validate_ephemeral_storage(self, ephemeral_storage):
        if not isinstance(ephemeral_storage, int):
            self._error(f'Ephemeral storage size must an integer but not '
                        f'\'{type(ephemeral_storage).__name__}\'')
        if not (512 <= ephemeral_storage <= 10240):
            self._error('Ephemeral storage size must be between '
                        '512 and 10240 MB')

    def _validate_architecture(self, architectures):
        for architecture in architectures:
            if architecture not in LAMBDA_ARCHITECTURE_LIST:
                self._error(f'Specified unsupported architecture: '
                            f'"{architecture}". Currently supported '
                            f'architectures: {LAMBDA_ARCHITECTURE_LIST}')


def validate_lambda(name, meta):
    LambdaValidator(name, meta).validate()



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.waiter import WaiterError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.batch_compenv')

DEFAULT_STATE = 'ENABLED'
DEFAULT_SERVICE_ROLE = 'AWSBatchServiceRole'


class BatchComputeEnvironmentResource(BaseResource):

    def __init__(self, batch_conn, iam_conn, region, account_id):
        self.batch_conn = batch_conn
        self.iam_conn = iam_conn
        self.region = region
        self.account_id = account_id

    def create_compute_environment(self, args):
        return self.create_pool(self._create_compute_environment_from_meta,
                                args)

    def describe_compute_environment(self, name, meta):
        response = self.batch_conn.describe_compute_environments(name)

        try:
            arn = response['computeEnvironments'][0]['computeEnvironmentArn']
            return {arn: build_description_obj(response, name, meta)}
        except (KeyError, IndexError):
            _LOG.warn("Batch Compute Environment %s not found", name)
            return {}

    def remove_compute_environment(self, args):
        self.create_pool(self._remove_compute_environment, args)

    @unpack_kwargs
    def _remove_compute_environment(self, arn, config):
        compute_environment_data = self.batch_conn.describe_compute_environments(arn)
        try:
            compute_environment_data = compute_environment_data[
                'computeEnvironments'][0]
        except (KeyError, IndexError):
            _LOG.warn("Batch Compute Environment %s not found", config[
                'resource_name'])
            return
        if compute_environment_data['state'] == 'ENABLED':
            # need to disable compute env first
            self.batch_conn.update_compute_environment(arn, state='DISABLED')

        self.batch_conn.delete_compute_environment(compute_environment=arn)

        compute_environment_name = compute_environment_data[
            'computeEnvironmentName']
        _LOG.info('Batch Compute Environment %s was removed.',
                  compute_environment_name)

    @unpack_kwargs
    def _create_compute_environment_from_meta(self, name, meta):
        from syndicate.core import CONFIG
        params = meta.copy()
        params['compute_environment_name'] = name

        if 'resource_type' in params:
            del params['resource_type']
        if self._is_compute_env_exist(name):
            _LOG.warn(f'AWS Batch Compute Environment with the name {name} '
                      f'already exists')
            return self.describe_compute_environment(name, meta)

        state = params.get('state')
        if not state:
            params['state'] = DEFAULT_STATE

        service_role = params.get('service_role')
        if not service_role:
            role = self.iam_conn.get_role(role_name=DEFAULT_SERVICE_ROLE)
            if not role:
                _LOG.warn("Default Service Role '%s' not found and will be "
                          "created", DEFAULT_SERVICE_ROLE)
                allowed_account = self.account_id
                self.iam_conn.create_custom_role(
                    role_name=DEFAULT_SERVICE_ROLE,
                    allowed_account=allowed_account,
                    allowed_service='batch',
                    permissions_boundary=CONFIG.iam_permissions_boundary
                )
                policy_arn = self.iam_conn.get_policy_arn(DEFAULT_SERVICE_ROLE)
                self.iam_conn.attach_policy(
                    role_name=DEFAULT_SERVICE_ROLE,
                    policy_arn=policy_arn
                )
                _LOG.info("Created default service role %s", DEFAULT_SERVICE_ROLE)
            params['service_role'] = DEFAULT_SERVICE_ROLE

        # resolve IAM Role name with IAM Role ARN
        params['service_role'] = self.iam_conn.check_if_role_exists(
            role_name=params['service_role'])

        self.batch_conn.create_compute_environment(**params)
        try:
            waiter = self.batch_conn.get_compute_environment_waiter()
            waiter.wait(computeEnvironments=[name])
        except WaiterError as e:
            _LOG.error(e)

        _LOG.info('Created Batch Compute Environment %s.', name)
        return self.describe_compute_environment(name, meta)

    def _is_compute_env_exist(self, compute_environment_name):
        response = self.batch_conn.describe_compute_environments(
            compute_environment_name)
        return bool(response['computeEnvironments'])

    def update_compute_environment(self, args):
        return self.create_pool(self._update_compute_environment_from_meta,
                                args)

    def _update_compute_environment_from_meta(self, meta):
        from syndicate.core import CONFIG
        name = meta.pop('name')
        arn = f'arn:aws:batch:{self.region}:{self.account_id}:' \
              f'compute-environment/{name}'
        if not self._is_compute_env_exist(arn):
            raise AssertionError(f'Compute environment \'{name}\' does not '
                                 f'exist')

        params = meta['meta'].copy()
        if 'resource_type' in params:
            del params['resource_type']
        if 'compute_environment_type' in params:
            del params['compute_environment_type']
        if 'subnets' in params:
            del params['subnets']
        if 'compute_resources' in params:
            del params['compute_resources']

        state = params.get('state')
        if state and state != 'ENABLED' and state != 'DISABLED':
            _LOG.warn(f"Invalid state value for compute environment '{arn}': "
                      f"{state}")
            raise AssertionError(f'Invalid state value for compute '
                                 f'environment \'{arn}\': {state}')

        params['compute_environment'] = arn

        service_role = params.get('service_role')
        if not service_role:
            role = self.iam_conn.get_role(role_name=DEFAULT_SERVICE_ROLE)
            if not role:
                _LOG.warn(f"Default Service Role '{DEFAULT_SERVICE_ROLE}' not "
                          f"found and will be created")
                allowed_account = self.account_id
                self.iam_conn.create_custom_role(
                    role_name=DEFAULT_SERVICE_ROLE,
                    allowed_account=allowed_account,
                    allowed_service='batch',
                    permissions_boundary=CONFIG.iam_permissions_boundary
                )
                policy_arn = self.iam_conn.get_policy_arn(DEFAULT_SERVICE_ROLE)
                self.iam_conn.attach_policy(
                    role_name=DEFAULT_SERVICE_ROLE,
                    policy_arn=policy_arn
                )
                _LOG.info(f"Created default service role "
                          f"{DEFAULT_SERVICE_ROLE}")
            params['service_role'] = DEFAULT_SERVICE_ROLE

        # resolve IAM Role name with IAM Role ARN
        params['service_role'] = self.iam_conn.check_if_role_exists(
            role_name=params['service_role'])

        # response: TypedDict[computeEnvironmentName|Arn, ResponseMetadata]
        _response: dict = self.batch_conn.update_compute_environment(**params)
        return self.describe_compute_environment(name=name, meta=meta)



import os

import boto3
import requests
from requests_aws_sign import AWSV4Sign

from syndicate.commons.log_helper import get_logger
from syndicate.core import ResourceProvider
from syndicate.core.build.bundle_processor import load_deploy_output
from syndicate.core.conf.processor import ConfigHolder
from syndicate.core.helper import exit_on_exception

_LOG = get_logger('syndicate.core.build.warmup_processor')

NODEJS_RUNTIME = 'nodejs'
PYTHON_RUNTIME = 'python'

GET_METHOD = 'GET'
POST_METHOD = 'POST'
PUT_METHOD = 'PUT'
PATCH_METHOD = 'PATCH'
DELETE_METHOD = 'DELETE'

methods_check = {
    POST_METHOD: requests.post,
    GET_METHOD: requests.get,
    PUT_METHOD: requests.put,
    PATCH_METHOD: requests.patch,
    DELETE_METHOD: requests.delete
}


def process_api_gw_resources(paths_to_be_triggered,
                             resource_path_warmup_key_mapping):
    resource_method_mapping = {}
    resource_warmup_key_mapping = {}
    for api_gw_link, path_method_mapping in paths_to_be_triggered.items():
        for path, method in path_method_mapping.items():
            resource_link = api_gw_link + path
            resource_method_mapping.update({resource_link: method})
            resource_warmup_key_mapping.update(
                {resource_link: resource_path_warmup_key_mapping[api_gw_link]
                [path]})
    return resource_method_mapping, resource_warmup_key_mapping


def get_aws_sign():
    config = __get_config()

    boto3_session = boto3.session.Session()
    credentials = boto3_session.get_credentials()
    credentials.access_key = config.aws_access_key_id
    credentials.secret_key = config.aws_secret_access_key
    region = config.region

    service = 'execute-api'
    auth = AWSV4Sign(credentials, region, service)
    return auth


def warm_upper(resource_method_mapping, resource_warmup_key_mapping,
               lambda_auth, header_name, header_value):
    for uri, methods in resource_method_mapping.items():
        for method in methods:
            warmup_method = methods_check.get(method)
            warmup_key = resource_warmup_key_mapping[uri]
            params = {warmup_key: "true"}
            if lambda_auth:
                headers = {header_name: header_value}
                warmup_method(uri, headers=headers, params=params)
            else:
                auth = get_aws_sign()
                warmup_method(uri, auth=auth, params=params)


def _get_api_gw_client():
    return ResourceProvider.instance.api_gw().connection.client


def get_api_stage(rest_api_id, user_input_stage_name, echo):
    api_gw_client = _get_api_gw_client()
    stages_info = api_gw_client.get_stages(restApiId=rest_api_id)
    stages = stages_info.get('item')
    all_stage_names = [stage.get('stageName') for stage in stages]

    if not user_input_stage_name:
        if len(all_stage_names) == 1:
            stage_name = all_stage_names[0]
        else:
            echo(f'Stage name(s) for {rest_api_id} API ID: '
                 f'{", ".join(all_stage_names)}')
            stage_name = input('Select Stage from existing: ')
            if stage_name not in all_stage_names:
                raise AssertionError(f'Provided Stage name does not exists')

    else:
        if isinstance(user_input_stage_name, str):
            stage_name = [user_input_stage_name] \
                if user_input_stage_name in all_stage_names else None
        else:
            stage_name = [stage for stage in user_input_stage_name
                          if stage in all_stage_names]
        if not stage_name:
            raise AssertionError(
                f'Provided Stage name does not exists, available stage '
                f'name(s): {", ".join(all_stage_names)}')
        stage_name = stage_name[0]

    return stage_name


def get_warmup_param(allowed_path_warmup_key_mapping, api_gw_path, lambda_arn):
    lambda_client = ResourceProvider.instance.lambda_resource().lambda_conn. \
        client

    lambda_name = lambda_arn.split('function:')[-1]
    func_name = lambda_name.split('/')[0]

    lambda_meta = lambda_client.get_function(FunctionName=func_name)
    lambda_runtime = lambda_meta['Configuration']['Runtime']

    warmup_param = 'warmUp'
    if lambda_runtime.startswith(PYTHON_RUNTIME) or \
            lambda_runtime.startswith(NODEJS_RUNTIME):
        warmup_param = 'warm_up'

    allowed_path_warmup_key_mapping.update({api_gw_path: warmup_param})
    return allowed_path_warmup_key_mapping


def get_api_gw_resources(api_gw_client, rest_api_id):
    resource = api_gw_client.get_resources(
        restApiId=rest_api_id,
        limit=500
    )
    resource_items = resource['items']
    resource_items_id = {item['id']: item['resourceMethods']
                         for item in resource_items
                         if 'resourceMethods' in item}

    resource_items_path = {item['id']: item['path']
                           for item in resource_items}
    return resource_items_path, resource_items_id


def get_api_gw_integration(rest_api_id):
    api_gw_client = _get_api_gw_client()
    resource_items_path, resource_items_id = get_api_gw_resources(
        api_gw_client, rest_api_id)

    affected_lambda = []
    allowed_path_method_mapping = {}
    allowed_path_warmup_key_mapping = {}
    for resource_id, methods in resource_items_id.items():
        for method in methods:
            integration = api_gw_client.get_integration(restApiId=rest_api_id,
                                                        resourceId=resource_id,
                                                        httpMethod=method)

            api_gw_path = resource_items_path[resource_id]
            lambda_arn = integration.get('uri')
            if lambda_arn and lambda_arn not in affected_lambda:
                affected_lambda.append(lambda_arn)

                if not api_gw_path in allowed_path_method_mapping:
                    allowed_path_method_mapping.update({api_gw_path: [method]})
                else:
                    allowed_path_method_mapping[api_gw_path].append(method)

                get_warmup_param(allowed_path_warmup_key_mapping, api_gw_path,
                                 lambda_arn)

    return allowed_path_method_mapping, allowed_path_warmup_key_mapping


def __get_config():
    conf_path = os.environ.get('SDCT_CONF')
    config = ConfigHolder(conf_path)
    return config


def get_api_gw_link(rest_api_id, stage_name):
    api_gw_link_template = 'https://{rest_api_id}.execute-api.{region}.' \
                           'amazonaws.com/{stage_name}'

    config = __get_config()
    region = config.region
    api_gw_link = api_gw_link_template.format(rest_api_id=rest_api_id,
                                              region=region,
                                              stage_name=stage_name)
    return api_gw_link


@exit_on_exception
def process_deploy_resources(bundle_name, deploy_name):
    output = load_deploy_output(bundle_name, deploy_name)

    output = {key: value for key, value in output.items() if
              value['resource_meta'].get('resource_type') == 'api_gateway'}

    paths_to_be_triggered = {}
    resource_path_warmup_key_mapping = {}
    if not output:
        _LOG.warning('No resources to warmup, exiting')
        return paths_to_be_triggered, resource_path_warmup_key_mapping

    for resource_arn, meta in output.items():
        rest_api_id = resource_arn.split('/')[-1]
        stage_name = meta.get('resource_meta', {}).get('deploy_stage')

        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            handle_paths_to_be_triggered(rest_api_id=rest_api_id,
                                         stage_name=stage_name,
                                         paths_to_be_triggered=
                                         paths_to_be_triggered,
                                         resource_path_warmup_key_mapping=
                                         resource_path_warmup_key_mapping)
    return paths_to_be_triggered, resource_path_warmup_key_mapping


def process_inputted_api_gw_id(api_id, stage_name, echo):
    api_gw_client = _get_api_gw_client()
    all_apis = api_gw_client.get_rest_apis().get('items', {})

    allowed_id = [api['id'] for api in all_apis]

    paths_to_be_triggered = {}
    resource_path_warmup_key_mapping = {}

    for rest_api_id in api_id:
        if rest_api_id not in allowed_id:
            echo(f'Provided {rest_api_id} API ID does not exists')
            continue

        stage_name = get_api_stage(rest_api_id, stage_name, echo=echo)
        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            handle_paths_to_be_triggered(rest_api_id=rest_api_id,
                                         stage_name=stage_name,
                                         paths_to_be_triggered=
                                         paths_to_be_triggered,
                                         resource_path_warmup_key_mapping=
                                         resource_path_warmup_key_mapping)
    return paths_to_be_triggered, resource_path_warmup_key_mapping


def process_existing_api_gw_id(stage_name, echo):
    api_gw_client = _get_api_gw_client()
    all_apis = api_gw_client.get_rest_apis().get('items', {})

    all_api_name = {api['name']: api['id'] for api in all_apis}

    echo(f'Existed API Gateway: {", ".join(all_api_name)}')
    user_input_id = input('Select API from existing (multiple names must be'
                          ' separated by commas): ')
    user_input_id = user_input_id.split(",")
    user_input_id = {user_id.strip() for user_id in user_input_id}

    if not user_input_id.issubset(all_api_name):
        raise AssertionError(
            f'Specify only allowed IDs: {", ".join(all_api_name)}')

    paths_to_be_triggered = {}
    resource_path_warmup_key_mapping = {}

    for api_name in user_input_id:
        rest_api_id = all_api_name[api_name]
        stage_name = get_api_stage(rest_api_id, stage_name, echo)
        handle_paths_to_be_triggered(rest_api_id, stage_name,
                                     paths_to_be_triggered,
                                     resource_path_warmup_key_mapping)
        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            handle_paths_to_be_triggered(rest_api_id=rest_api_id,
                                         stage_name=stage_name,
                                         paths_to_be_triggered=
                                         paths_to_be_triggered,
                                         resource_path_warmup_key_mapping=
                                         resource_path_warmup_key_mapping)
    return paths_to_be_triggered, resource_path_warmup_key_mapping


def handle_paths_to_be_triggered(rest_api_id, stage_name,
                                 paths_to_be_triggered,
                                 resource_path_warmup_key_mapping):
    api_gw_link = get_api_gw_link(rest_api_id, stage_name)
    allowed_path_method_mapping, allowed_path_warmup_key_mapping = \
        get_api_gw_integration(rest_api_id)
    paths_to_be_triggered.update(
        {api_gw_link: allowed_path_method_mapping})
    resource_path_warmup_key_mapping.update(
        {api_gw_link: allowed_path_warmup_key_mapping})
    return paths_to_be_triggered, resource_path_warmup_key_mapping



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

PROVIDER_KEY = 'provider'
RESOURCE_KEY = 'resource'

LAMBDA_RESOURCE_NAME = 'aws_lambda_function'
IAM_POLICY_RESOURCE_NAME = 'aws_iam_policy'
IAM_ROLE_RESOURCE_NAME = 'aws_iam_role'
DYNAMO_DB_TABLE_RESOURCE_NAME = 'aws_dynamodb_table'
APP_AUTOSCALING_TARGET_RESOURCE_NAME = 'aws_appautoscaling_target'
APP_AUTOSCALING_POLICY_RESOURCE_NAME = 'aws_appautoscaling_policy'
API_GATEWAY_REST_API_RESOURCE_NAME = 'aws_api_gateway_rest_api'
API_GATEWAY_RESOURCE_RESOURCE_NAME = 'aws_api_gateway_resource'
API_GATEWAY_METHOD_RESPONSE_RESOURCE_NAME = 'aws_api_gateway_method_response'
API_GATEWAY_INTEGRATION_RESOURCE_NAME = 'aws_api_gateway_integration'
API_GATEWAY_STAGE_RESOURCE_NAME = 'aws_api_gateway_stage'
API_GATEWAY_DEPLOYMENT_RESOURCE_NAME = 'aws_api_gateway_deployment'
API_GATEWAY_INTEGRATION_RESPONSE_RESOURCE_NAME = 'aws_api_gateway_integration_response'
API_GATEWAY_METHOD_RESOURCE_NAME = 'aws_api_gateway_method'
S3_BUCKET_RESOURCE_NAME = 'aws_s3_bucket'
CLOUD_WATCH_EVENT_RULE = 'aws_cloudwatch_event_rule'
CLOUD_WATCH_EVENT_TARGET = 'aws_cloudwatch_event_target'
SNS_TOPIC_RESOURCE = 'aws_sns_topic'
SQS_QUEUE_RESOURCE = 'aws_sqs_queue'
CLOUD_WATCH_ALARM = 'aws_cloudwatch_metric_alarm'
KINESIS_STREAM_RESOURCE = 'aws_kinesis_stream'
SNS_APPLICATION_RESOURCE = 'aws_sns_platform_application'
AWS_API_GATEWAY_AUTHORIZER = 'aws_api_gateway_authorizer'
AWS_COGNITO_IDENTITY_POOL = 'aws_cognito_identity_pool'
AWS_COGNITO_IDENTITY_POOL_ROLES_ATTACHMENT = 'aws_cognito_identity_pool_roles_attachment'
AWS_LAMBDA_PROVISIONED_CONCURRENCY_CONFIG = 'aws_lambda_provisioned_concurrency_config'
AWS_LAMBDA_ALIAS = 'aws_lambda_alias'
CLOUD_WATCH_LOG_GROUP = 'aws_cloudwatch_log_group'
SNS_TOPIC_POLICY = 'aws_sns_topic_policy'
AWS_BATCH_JOB_DEFINITION = 'aws_batch_job_definition'
AWS_BATCH_COMPUTE_ENVIRONMENT = 'aws_batch_compute_environment'
AWS_BATCH_JOB_QUEUE = 'aws_batch_job_queue'
AWS_IAM_INSTANCE_PROFILE = 'aws_iam_instance_profile'
AWS_IAM_ROLE_POLICY_ATTACHMENT = 'aws_iam_role_policy_attachment'
AWS_LAMBDA_EVENT_SOURCE_MAPPING = 'aws_lambda_event_source_mapping'
AWS_IAM_ROLE_POLICY = 'aws_iam_role_policy'
AWS_SNS_TOPIC_SUBSCRIPTION = 'aws_sns_topic_subscription'
AWS_LAMBDA_PERMISSION = 'aws_lambda_permission'
AWS_S3_BUCKET_NOTIFICATION = 'aws_s3_bucket_notification'
AWS_SQS_POLICY = 'aws_sqs_queue_policy'
AWS_API_GATEWAY_REQUEST_VALIDATOR = 'aws_api_gateway_request_validator'
AWS_LAMBDA_LAYER_VERSION = 'aws_lambda_layer_version'

RESOURCE_TYPES = [LAMBDA_RESOURCE_NAME, IAM_POLICY_RESOURCE_NAME,
                  IAM_ROLE_RESOURCE_NAME, DYNAMO_DB_TABLE_RESOURCE_NAME,
                  APP_AUTOSCALING_TARGET_RESOURCE_NAME,
                  APP_AUTOSCALING_POLICY_RESOURCE_NAME,
                  API_GATEWAY_REST_API_RESOURCE_NAME,
                  API_GATEWAY_RESOURCE_RESOURCE_NAME,
                  API_GATEWAY_METHOD_RESPONSE_RESOURCE_NAME,
                  API_GATEWAY_INTEGRATION_RESOURCE_NAME,
                  API_GATEWAY_STAGE_RESOURCE_NAME,
                  API_GATEWAY_DEPLOYMENT_RESOURCE_NAME,
                  API_GATEWAY_INTEGRATION_RESPONSE_RESOURCE_NAME,
                  API_GATEWAY_METHOD_RESOURCE_NAME, S3_BUCKET_RESOURCE_NAME,
                  CLOUD_WATCH_EVENT_RULE, CLOUD_WATCH_EVENT_TARGET,
                  SNS_TOPIC_RESOURCE, SQS_QUEUE_RESOURCE, CLOUD_WATCH_ALARM,
                  KINESIS_STREAM_RESOURCE, SNS_APPLICATION_RESOURCE,
                  AWS_API_GATEWAY_AUTHORIZER, AWS_COGNITO_IDENTITY_POOL,
                  AWS_COGNITO_IDENTITY_POOL_ROLES_ATTACHMENT,
                  AWS_LAMBDA_PROVISIONED_CONCURRENCY_CONFIG, AWS_LAMBDA_ALIAS,
                  CLOUD_WATCH_LOG_GROUP, SNS_TOPIC_POLICY,
                  AWS_BATCH_JOB_DEFINITION, AWS_BATCH_COMPUTE_ENVIRONMENT,
                  AWS_BATCH_JOB_QUEUE, AWS_IAM_INSTANCE_PROFILE,
                  AWS_IAM_ROLE_POLICY_ATTACHMENT,
                  AWS_LAMBDA_EVENT_SOURCE_MAPPING,
                  AWS_IAM_ROLE_POLICY, AWS_SNS_TOPIC_SUBSCRIPTION,
                  AWS_LAMBDA_PERMISSION, AWS_S3_BUCKET_NOTIFICATION,
                  AWS_SQS_POLICY, AWS_API_GATEWAY_REQUEST_VALIDATOR,
                  AWS_LAMBDA_LAYER_VERSION]


class TerraformTemplate(object):

    def __init__(self, provider, profile, region):
        self.aws_lambda_function = []
        self.aws_iam_policy = []
        self.aws_iam_role = []
        self.aws_dynamodb_table = []
        self.aws_appautoscaling_target = []
        self.aws_appautoscaling_policy = []
        self.aws_api_gateway_rest_api = []
        self.aws_api_gateway_resource = []
        self.aws_api_gateway_method_response = []
        self.aws_api_gateway_integration = []
        self.aws_api_gateway_stage = []
        self.aws_api_gateway_deployment = []
        self.aws_api_gateway_integration_response = []
        self.aws_api_gateway_method = []
        self.aws_s3_bucket = []
        self.aws_cloudwatch_event_rule = []
        self.aws_cloudwatch_event_target = []
        self.aws_sns_topic = []
        self.aws_sqs_queue = []
        self.aws_cloudwatch_metric_alarm = []
        self.aws_kinesis_stream = []
        self.aws_sns_platform_application = []
        self.aws_api_gateway_authorizer = []
        self.aws_cognito_identity_pool = []
        self.aws_cognito_identity_pool_roles_attachment = []
        self.aws_lambda_provisioned_concurrency_config = []
        self.aws_lambda_alias = []
        self.cloud_watch_log_group = []
        self.aws_sns_topic_policy = []
        self.aws_batch_job_definition = []
        self.aws_batch_compute_environment = []
        self.aws_batch_job_queue = []
        self.aws_iam_instance_profile = []
        self.aws_iam_role_policy_attachment = []
        self.aws_lambda_event_source_mapping = []
        self.aws_iam_role_policy = []
        self.aws_sns_topic_subscription = []
        self.aws_lambda_permission = []
        self.aws_s3_bucket_notification = []
        self.aws_sqs_queue_policy = []
        self.aws_api_gateway_request_validator = []
        self.aws_lambda_layer_version = []

        self.compose_resources_mapping = {
            LAMBDA_RESOURCE_NAME: self._aws_lambda,
            IAM_POLICY_RESOURCE_NAME: self._aws_iam_policy,
            IAM_ROLE_RESOURCE_NAME: self._aws_iam_role,
            DYNAMO_DB_TABLE_RESOURCE_NAME: self._aws_dynamodb_table,
            APP_AUTOSCALING_TARGET_RESOURCE_NAME: self._aws_appautoscaling_target,
            APP_AUTOSCALING_POLICY_RESOURCE_NAME: self._aws_appautoscaling_policy,
            API_GATEWAY_REST_API_RESOURCE_NAME: self._aws_api_gateway_rest_api,
            API_GATEWAY_RESOURCE_RESOURCE_NAME: self._aws_api_gateway_resource,
            API_GATEWAY_METHOD_RESPONSE_RESOURCE_NAME: self._aws_api_gateway_method_response,
            API_GATEWAY_INTEGRATION_RESOURCE_NAME: self._aws_api_gateway_integration,
            API_GATEWAY_STAGE_RESOURCE_NAME: self._aws_api_gateway_stage,
            API_GATEWAY_DEPLOYMENT_RESOURCE_NAME: self._aws_api_gateway_deployment,
            API_GATEWAY_INTEGRATION_RESPONSE_RESOURCE_NAME: self._aws_api_gateway_integration_response,
            API_GATEWAY_METHOD_RESOURCE_NAME: self._aws_api_gateway_method,
            S3_BUCKET_RESOURCE_NAME: self._aws_aws_s3_bucket,
            CLOUD_WATCH_EVENT_RULE: self._aws_cloudwatch_event_rule,
            CLOUD_WATCH_EVENT_TARGET: self._aws_cloudwatch_event_target,
            SNS_TOPIC_RESOURCE: self._aws_sns_topic,
            SQS_QUEUE_RESOURCE: self._aws_sqs_queue,
            CLOUD_WATCH_ALARM: self._aws_cloudwatch_metric_alarm,
            KINESIS_STREAM_RESOURCE: self._aws_kinesis_stream,
            SNS_APPLICATION_RESOURCE: self._aws_sns_platform_application,
            AWS_API_GATEWAY_AUTHORIZER: self._aws_api_gateway_authorizer,
            AWS_COGNITO_IDENTITY_POOL: self._aws_cognito_identity_pool,
            AWS_COGNITO_IDENTITY_POOL_ROLES_ATTACHMENT: self._aws_cognito_identity_pool_roles_attachment,
            AWS_LAMBDA_PROVISIONED_CONCURRENCY_CONFIG: self._aws_lambda_provisioned_concurrency_config,
            AWS_LAMBDA_ALIAS: self._aws_lambda_alias,
            CLOUD_WATCH_LOG_GROUP: self._cloud_watch_log_group,
            SNS_TOPIC_POLICY: self._aws_sns_topic_policy,
            AWS_BATCH_JOB_DEFINITION: self._aws_batch_job_definition,
            AWS_BATCH_COMPUTE_ENVIRONMENT: self._aws_batch_compute_environment,
            AWS_BATCH_JOB_QUEUE: self._aws_batch_job_queue,
            AWS_IAM_INSTANCE_PROFILE: self._aws_iam_instance_profile,
            AWS_IAM_ROLE_POLICY_ATTACHMENT: self._aws_iam_role_policy_attachment,
            AWS_LAMBDA_EVENT_SOURCE_MAPPING: self._aws_lambda_event_source_mapping,
            AWS_IAM_ROLE_POLICY: self._aws_iam_role_policy,
            AWS_SNS_TOPIC_SUBSCRIPTION: self._aws_sns_topic_subscription,
            AWS_LAMBDA_PERMISSION: self._aws_lambda_permission,
            AWS_S3_BUCKET_NOTIFICATION: self._aws_s3_bucket_notification,
            AWS_SQS_POLICY: self._aws_sqs_queue_policy,
            AWS_API_GATEWAY_REQUEST_VALIDATOR: self._aws_api_gateway_request_validator,
            AWS_LAMBDA_LAYER_VERSION: self._aws_lambda_layer_version
        }

        self.provider = provider
        self.resources = list()
        self.providers = list()
        default_provider = {
            "profile": profile,
            "region": region
        }
        self.providers.append(default_provider)

    def add_aws_lambda(self, meta):
        self.aws_lambda_function.append(meta)

    def add_aws_iam_policy(self, meta):
        self.aws_iam_policy.append(meta)

    def add_aws_iam_role(self, meta):
        self.aws_iam_role.append(meta)

    def add_aws_dynamodb_table(self, meta):
        self.aws_dynamodb_table.append(meta)

    def add_aws_appautoscaling_target(self, meta):
        self.aws_appautoscaling_target.append(meta)

    def add_aws_appautoscaling_policy(self, meta):
        self.aws_appautoscaling_policy.append(meta)

    def add_aws_api_gateway_rest_api(self, meta):
        self.aws_api_gateway_rest_api.append(meta)

    def add_aws_api_gateway_resource(self, meta):
        self.aws_api_gateway_resource.append(meta)

    def add_aws_api_gateway_method_response(self, meta):
        self.aws_api_gateway_method_response.append(meta)

    def add_aws_api_gateway_integration(self, meta):
        self.aws_api_gateway_integration.append(meta)

    def add_aws_api_gateway_stage(self, meta):
        self.aws_api_gateway_stage.append(meta)

    def add_aws_api_gateway_deployment(self, meta):
        self.aws_api_gateway_deployment.append(meta)

    def add_aws_api_gateway_integration_response(self, meta):
        self.aws_api_gateway_integration_response.append(meta)

    def add_aws_api_gateway_method(self, meta):
        self.aws_api_gateway_method.append(meta)

    def add_aws_s3_bucket(self, meta):
        self.aws_s3_bucket.append(meta)

    def add_aws_cloudwatch_event_rule(self, meta):
        self.aws_cloudwatch_event_rule.append(meta)

    def add_aws_batch_job_definition(self, meta):
        self.aws_batch_job_definition.append(meta)

    def add_aws_cloudwatch_event_target(self, meta):
        self.aws_cloudwatch_event_target.append(meta)

    def add_aws_sns_topic(self, meta):
        self.aws_sns_topic.append(meta)

    def add_aws_sqs_queue(self, meta):
        self.aws_sqs_queue.append(meta)

    def add_aws_cloudwatch_metric_alarm(self, meta):
        self.aws_cloudwatch_metric_alarm.append(meta)

    def add_aws_kinesis_stream(self, meta):
        self.aws_kinesis_stream.append(meta)

    def add_aws_sns_platform_application(self, meta):
        self.aws_sns_platform_application.append(meta)

    def add_aws_api_gateway_authorizer(self, meta):
        self.aws_api_gateway_authorizer.append(meta)

    def add_aws_cognito_identity_pool(self, meta):
        self.aws_cognito_identity_pool.append(meta)

    def add_aws_cognito_identity_pool_roles_attachment(self, meta):
        self.aws_cognito_identity_pool_roles_attachment.append(meta)

    def add_aws_lambda_provisioned_concurrency_config(self, meta):
        self.aws_lambda_provisioned_concurrency_config.append(meta)

    def add_aws_lambda_alias(self, meta):
        self.aws_lambda_alias.append(meta)

    def add_cloud_watch_log_group(self, meta):
        self.cloud_watch_log_group.append(meta)

    def add_aws_sns_topic_policy(self, meta):
        self.aws_sns_topic_policy.append(meta)

    def add_aws_batch_job_queue(self, meta):
        self.aws_batch_job_queue.append(meta)

    def add_aws_batch_job_definition(self, meta):
        self.aws_batch_job_definition.append(meta)

    def add_aws_batch_compute_environment(self, meta):
        self.aws_batch_compute_environment.append(meta)

    def add_aws_iam_instance_profile(self, meta):
        self.aws_iam_instance_profile.append(meta)

    def add_aws_iam_role_policy_attachment(self, meta):
        self.aws_iam_role_policy_attachment.append(meta)

    def add_aws_lambda_event_source_mapping(self, meta):
        self.aws_lambda_event_source_mapping.append(meta)

    def add_aws_iam_role_policy(self, meta):
        self.aws_iam_role_policy.append(meta)

    def add_aws_sns_topic_subscription(self, meta):
        self.aws_sns_topic_subscription.append(meta)

    def add_aws_lambda_permission(self, meta):
        self.aws_lambda_permission.append(meta)

    def add_aws_s3_bucket_notification(self, meta):
        self.aws_s3_bucket_notification.append(meta)

    def add_aws_sqs_queue_policy(self, meta):
        self.aws_sqs_queue_policy.append(meta)

    def add_aws_api_gateway_request_validator(self, meta):
        self.aws_api_gateway_request_validator.append(meta)

    def add_aws_lambda_layer_version(self, meta):
        self.aws_lambda_layer_version.append(meta)

    def add_provider_if_not_exists(self, region):
        for provider in self.providers:
            if provider['region'] == region:
                return
        self.providers.append(
            {
                'region': region,
                'alias': region
            }
        )

    def _aws_lambda(self):
        return self.aws_lambda_function

    def _aws_iam_policy(self):
        return self.aws_iam_policy

    def _aws_iam_role(self):
        return self.aws_iam_role

    def _aws_dynamodb_table(self):
        return self.aws_dynamodb_table

    def _aws_appautoscaling_target(self):
        return self.aws_appautoscaling_target

    def _aws_appautoscaling_policy(self):
        return self.aws_appautoscaling_policy

    def _aws_api_gateway_rest_api(self):
        return self.aws_api_gateway_rest_api

    def _aws_api_gateway_resource(self):
        return self.aws_api_gateway_resource

    def _aws_api_gateway_method_response(self):
        return self.aws_api_gateway_method_response

    def _aws_api_gateway_integration(self):
        return self.aws_api_gateway_integration

    def _aws_api_gateway_stage(self):
        return self.aws_api_gateway_stage

    def _aws_api_gateway_deployment(self):
        return self.aws_api_gateway_deployment

    def _aws_api_gateway_integration_response(self):
        return self.aws_api_gateway_integration_response

    def _aws_api_gateway_method(self):
        return self.aws_api_gateway_method

    def _aws_aws_s3_bucket(self):
        return self.aws_s3_bucket

    def _aws_cloudwatch_event_rule(self):
        return self.aws_cloudwatch_event_rule

    def _aws_cloudwatch_event_target(self):
        return self.aws_cloudwatch_event_target

    def _aws_sns_topic(self):
        return self.aws_sns_topic

    def _aws_sqs_queue(self):
        return self.aws_sqs_queue

    def _aws_cloudwatch_metric_alarm(self):
        return self.aws_cloudwatch_metric_alarm

    def _aws_kinesis_stream(self):
        return self.aws_kinesis_stream

    def _aws_sns_platform_application(self):
        return self.aws_sns_platform_application

    def _aws_api_gateway_authorizer(self):
        return self.aws_api_gateway_authorizer

    def _aws_cognito_identity_pool(self):
        return self.aws_cognito_identity_pool

    def _aws_cognito_identity_pool_roles_attachment(self):
        return self.aws_cognito_identity_pool_roles_attachment

    def _aws_lambda_provisioned_concurrency_config(self):
        return self.aws_lambda_provisioned_concurrency_config

    def _aws_lambda_alias(self):
        return self.aws_lambda_alias

    def _cloud_watch_log_group(self):
        return self.cloud_watch_log_group

    def _aws_sns_topic_policy(self):
        return self.aws_sns_topic_policy

    def _aws_batch_job_definition(self):
        return self.aws_batch_job_definition

    def _aws_batch_compute_environment(self):
        return self.aws_batch_compute_environment

    def _aws_batch_job_queue(self):
        return self.aws_batch_job_queue

    def _aws_iam_instance_profile(self):
        return self.aws_iam_instance_profile

    def _aws_iam_role_policy_attachment(self):
        return self.aws_iam_role_policy_attachment

    def _aws_lambda_event_source_mapping(self):
        return self.aws_lambda_event_source_mapping

    def _aws_iam_role_policy(self):
        return self.aws_iam_role_policy

    def _aws_sns_topic_subscription(self):
        return self.aws_sns_topic_subscription

    def _aws_lambda_permission(self):
        return self.aws_lambda_permission

    def _aws_s3_bucket_notification(self):
        return self.aws_s3_bucket_notification

    def _aws_sqs_queue_policy(self):
        return self.aws_sqs_queue_policy

    def _aws_api_gateway_request_validator(self):
        return self.aws_api_gateway_request_validator

    def _aws_lambda_layer_version(self):
        return self.aws_lambda_layer_version

    def provider_type(self):
        return self.provider

    def get_resource_by_name(self, resource_name):
        for res_type in RESOURCE_TYPES:
            resource_extractor = self.compose_resources_mapping.get(res_type)
            resources_meta = resource_extractor()
            if resources_meta:
                for resource in resources_meta:
                    if resource.get(resource_name):
                        return resource.get(resource_name)

    def compose_resources(self):
        for res_type in RESOURCE_TYPES:
            resource_extractor = self.compose_resources_mapping.get(res_type)
            resources_meta = resource_extractor()
            if resources_meta:
                self.resources.append({res_type: resources_meta})
        return json.dumps({PROVIDER_KEY: {self.provider: self.providers},
                           RESOURCE_KEY: self.resources}, indent=2)



import click
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import BATCH_JOBQUEUE_TYPE, BATCH_COMPENV_TYPE
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.resources.batch_jobqueue_resource import DEFAULT_STATE

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.batch_jobqueue')
USER_LOG = get_user_logger()


class BatchJobqueueGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = BATCH_JOBQUEUE_TYPE
    CONFIGURATION = {
        "state": DEFAULT_STATE,
        "priority": 1,
        "compute_environment_order": list,
    }

    def __init__(self, **kwargs):
        self.compute_environment_order = \
            kwargs.pop('compute_environment_order') or []
        super().__init__(**kwargs)

    def _generate_resource_configuration(self) -> dict:
        result = super()._generate_resource_configuration()
        for order, compute_env in self.compute_environment_order:
            paths_with_compute_env = self._get_resource_meta_paths(
                compute_env, BATCH_COMPENV_TYPE
            )
            if not paths_with_compute_env:
                message = f"Compute env '{compute_env}' hasn't been declared " \
                          f"in deployment resources yet."
                _LOG.warn(message)
                if click.confirm(
                        f"{message} Write compute env order anyway?"):
                    USER_LOG.warn(f"Writing compute env order "
                                  f"despite not declared compute env "
                                  f"'{compute_env}'")
                else:
                    USER_LOG.warn(
                        f"Skipping compute env in order '{compute_env}'...")
                    continue
            result['compute_environment_order'].append(
                {'order': order, 'compute_environment': compute_env})
        return result



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import applicationautoscaling as app_scaling
from troposphere import dynamodb

from syndicate.commons.log_helper import get_logger
from syndicate.connection.dynamo_connection import (
    DEFAULT_READ_CAPACITY, DEFAULT_WRITE_CAPACITY)
from syndicate.core.resources.dynamo_db_resource import (
    DYNAMODB_TABLE_REQUIRED_PARAMS, AUTOSCALING_REQUIRED_PARAMS,
    DynamoDBResource)
from syndicate.core.resources.helper import validate_params
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import (to_logic_name, iam_role_logic_name,
                                  dynamodb_table_logic_name)

_LOG = get_logger('cf_dynamodb_table_converter')


def _add_index_keys_to_definition(definition, indexes):
    if indexes:
        for index in indexes:
            _append_attr_definition(definition, index["index_key_name"],
                                    index["index_key_type"])
            if index.get('index_sort_key_name'):
                _append_attr_definition(definition,
                                        index["index_sort_key_name"],
                                        index["index_sort_key_type"])


def _append_attr_definition(definition, attr_name, attr_type):
    """
    :type definition: []
    :type attr_name: str
    :type attr_type: str
    """
    for each in definition:
        if each.AttributeName == attr_name:
            return
    definition.append(dynamodb.AttributeDefinition(
        AttributeName=attr_name,
        AttributeType=attr_type))


def _build_index_definition(index_meta, supplier):
    index_def = supplier(
        IndexName=index_meta['name'],
        KeySchema=[dynamodb.KeySchema(
            AttributeName=index_meta['index_key_name'],
            KeyType='HASH'
        )],
        Projection=dynamodb.Projection(
            ProjectionType='ALL'
        )
    )
    if index_meta.get('index_sort_key_name'):
        index_def.KeySchema.append(dynamodb.KeySchema(
            AttributeName=index_meta['index_sort_key_name'],
            KeyType='RANGE'))
    return index_def


class CfDynamoDbTableConverter(CfResourceConverter):

    def convert(self, name, meta):
        validate_params(name, meta, DYNAMODB_TABLE_REQUIRED_PARAMS)

        read_capacity = meta.get('read_capacity', DEFAULT_READ_CAPACITY)
        write_capacity = meta.get('write_capacity', DEFAULT_WRITE_CAPACITY)
        hash_key_name = meta['hash_key_name']
        hash_key_type = meta['hash_key_type']
        sort_key_name = meta.get('sort_key_name')

        schema = [dynamodb.KeySchema(AttributeName=hash_key_name,
                                     KeyType='HASH')]
        definition = [dynamodb.AttributeDefinition(
            AttributeName=hash_key_name,
            AttributeType=hash_key_type)]
        if sort_key_name:
            sort_key_type = meta['sort_key_type']
            schema.append(dynamodb.KeySchema(AttributeName=sort_key_name,
                                             KeyType='RANGE'))
            definition.append(dynamodb.AttributeDefinition(
                AttributeName=sort_key_name,
                AttributeType=sort_key_type))

        global_indexes = meta.get('global_indexes')
        local_indexes = meta.get('local_indexes')

        _add_index_keys_to_definition(definition=definition,
                                      indexes=global_indexes)
        _add_index_keys_to_definition(definition=definition,
                                      indexes=local_indexes)

        table = dynamodb.Table(dynamodb_table_logic_name(name))
        table.AttributeDefinitions = definition
        table.KeySchema = schema
        table.ProvisionedThroughput = dynamodb.ProvisionedThroughput(
            ReadCapacityUnits=read_capacity,
            WriteCapacityUnits=write_capacity)
        table.TableName = name
        self.template.add_resource(table)

        if global_indexes:
            global_secondary_indexes = []
            for index in global_indexes:
                index_def = _build_index_definition(
                    index_meta=index, supplier=dynamodb.GlobalSecondaryIndex)
                index_def.ProvisionedThroughput = \
                    dynamodb.ProvisionedThroughput(
                        ReadCapacityUnits=read_capacity,
                        WriteCapacityUnits=write_capacity)
                global_secondary_indexes.append(index_def)
            table.GlobalSecondaryIndexes = global_secondary_indexes

        if local_indexes:
            local_secondary_indexes = []
            for index in local_indexes:
                index_def = _build_index_definition(
                    index_meta=index, supplier=dynamodb.LocalSecondaryIndex)
                local_secondary_indexes.append(index_def)
            table.LocalSecondaryIndexes = local_secondary_indexes

        stream_view_type = meta.get('stream_view_type')
        self.configure_table_stream(table=table,
                                    stream_view_type=stream_view_type)

        autoscaling_config = meta.get('autoscaling')
        if autoscaling_config:
            for item in autoscaling_config:
                validate_params(name, item, AUTOSCALING_REQUIRED_PARAMS)
                role_name = item['role_name']
                role_res = self.get_resource(iam_role_logic_name(role_name))
                if role_res:
                    dimension = item['dimension']
                    resource_name = item['resource_name']

                    resource_id = DynamoDBResource.build_res_id(
                        dimension=dimension,
                        resource_name=resource_name,
                        table_name=name)
                    scalable_target = app_scaling.ScalableTarget(
                        to_logic_name(
                            'ApplicationAutoScalingScalableTarget',
                            resource_name))
                    scalable_target.MaxCapacity = str(item['max_capacity'])
                    scalable_target.MinCapacity = str(item['min_capacity'])
                    scalable_target.ResourceId = resource_id
                    scalable_target.RoleARN = role_res.get_att('Arn')
                    scalable_target.ScalableDimension = dimension
                    scalable_target.ServiceNamespace = 'dynamodb'
                    scalable_target.DependsOn = table
                    self.template.add_resource(scalable_target)

                    autoscaling_policy = item.get('config')
                    if autoscaling_policy:
                        policy_name = autoscaling_policy['policy_name']
                        target_value = autoscaling_policy['target_utilization']
                        scale_in_cooldown = \
                            autoscaling_policy.get('scale_in_cooldown')
                        scale_out_cooldown = \
                            autoscaling_policy.get('scale_out_cooldown')
                        metric_type = 'DynamoDBWriteCapacityUtilization' \
                            if 'Write' in dimension \
                            else 'DynamoDBReadCapacityUtilization'

                        scaling_policy = app_scaling.ScalingPolicy(
                            to_logic_name(
                                'ApplicationAutoScalingScalingPolicy',
                                resource_name))
                        scaling_policy.PolicyName = policy_name
                        scaling_policy.PolicyType = 'TargetTrackingScaling'
                        scaling_policy.ResourceId = resource_id
                        scaling_policy.ScalableDimension = dimension
                        scaling_policy.ServiceNamespace = 'dynamodb'
                        scaling_policy.DependsOn = scalable_target

                        metric_spec = app_scaling.PredefinedMetricSpecification(
                            PredefinedMetricType=metric_type)
                        _params = {
                            'PredefinedMetricSpecification': metric_spec,
                            'ScaleInCooldown': scale_in_cooldown,
                            'ScaleOutCooldown': scale_out_cooldown,
                            'TargetValue':target_value
                        }
                        _params = {k: v for k, v in _params.items() if
                                   isinstance(v, (bool, int)) or v}
                        scaling_policy.TargetTrackingScalingPolicyConfiguration = \
                            app_scaling.TargetTrackingScalingPolicyConfiguration(
                                **_params
                            )
                        self.template.add_resource(scaling_policy)
                else:
                    _LOG.warn('Role {} is not found in build meta, '
                              'skipping autoscaling config.'.format(role_name))

    @staticmethod
    def is_stream_enabled(table):
        try:
            stream = table.StreamSpecification
            return True
        except AttributeError:
            return False

    @staticmethod
    def configure_table_stream(table, stream_view_type='NEW_AND_OLD_IMAGES'):
        if stream_view_type:
            table.StreamSpecification = dynamodb.StreamSpecification(
                StreamViewType=stream_view_type)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from json import dumps
from functools import lru_cache

from boto3 import client, resource
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.iam_connection')


def get_account_role_arn(account_number):
    return "arn:aws:iam::{0}:root".format(account_number)


@apply_methods_decorator(retry())
class IAMConnection(object):
    """ IAM connection class."""

    def build_role_arn(self, role_name: str) -> str:
        from syndicate.core import CONFIG
        return f'arn:aws:iam::{CONFIG.account_id}:role' \
               f'/{CONFIG.resources_prefix}{role_name}{CONFIG.resources_suffix}'

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('iam', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        self.resource = resource('iam', region,
                                 aws_access_key_id=aws_access_key_id,
                                 aws_secret_access_key=aws_secret_access_key,
                                 aws_session_token=aws_session_token)
        _LOG.debug('Opened new IAM connection.')

    def check_if_role_exists(self, role_name):
        list_roles = self.get_roles()
        if list_roles:
            for each in list_roles:
                if role_name == each['RoleName']:
                    return each['Arn']

    def get_role(self, role_name):
        try:
            return self.client.get_role(RoleName=role_name)['Role']
        except ClientError as e:
            if 'NoSuchEntity' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_missing_roles(self, required_roles):
        list_roles = self.get_roles()
        if list_roles:
            all_roles = [role['RoleName'] for role in list_roles]
            return [role for role in required_roles if role not in all_roles]
        else:
            return required_roles

    def get_roles(self):
        roles = []
        response = self.client.list_roles()
        token = response.get('Marker')
        roles.extend(response.get('Roles'))
        while token:
            response = self.client.list_roles(Marker=token)
            token = response.get('Marker')
            roles.extend(response.get('Roles'))
        return roles

    @lru_cache()
    def get_policies(self, scope='All', only_attached=False):
        """
        :param scope: 'All'|'AWS'|'Local'
        :type only_attached: bool
        """
        policies = []
        response = self.client.list_policies(
            Scope=scope,
            OnlyAttached=only_attached
        )
        token = response.get('Marker')
        policies.extend(response.get('Policies'))
        while token:
            response = self.client.list_policies(
                Scope=scope,
                OnlyAttached=only_attached,
                Marker=token
            )
            token = response.get('Marker')
            policies.extend(response.get('Policies'))
        return policies

    def get_role_attached_policies(self, role_name):
        role = self.resource.Role(role_name)
        return role.attached_policies.all()

    def get_role_inline_policies(self, role_name):
        role = self.resource.Role(role_name)
        return role.policies.all()

    def get_role_names(self):
        roles = self.get_roles()
        return [each['RoleName'] for each in roles]

    def get_attached_policy_content(self, policy_arn):
        """ Get content from policy: latest version.

        :type policy_arn: str
        """
        version = self.client.list_policy_versions(PolicyArn=policy_arn)
        policy_version = version['Versions'][0]['VersionId']
        policy_content = self.client.get_policy_version(
            PolicyArn=policy_arn, VersionId=policy_version)
        return policy_content['PolicyVersion']['Document']

    def create_custom_policy(self, policy_name, policy_document):
        """
        :type policy_name: str
        :type policy_document: dict or str
        """
        if isinstance(policy_document, dict):
            policy_document = dumps(policy_document)
        return self.client.create_policy(
            PolicyName=policy_name,
            PolicyDocument=policy_document,
        )['Policy']

    def create_custom_role(self, role_name, allowed_account=None,
                           allowed_service=None, trusted_relationships=None,
                           external_id=None, permissions_boundary=None):
        """ Create custom role with trusted relationships. You can specify
        custom policy, or set principal_account and principal_service params
        to use default.

        :type role_name: str
        :type allowed_account: str (acc id)
        :type allowed_service: str
        :type trusted_relationships: dict
        :param trusted_relationships: if not specified will use default
        """
        if not trusted_relationships:
            trusted_relationships = IAMConnection.empty_trusted_relationships()
        if allowed_account:
            trusted_accounts = IAMConnection.set_allowed_account(
                allowed_account, external_id, 'create')
            trusted_relationships['Statement'].append(trusted_accounts)
        if allowed_service:
            trusted_services = IAMConnection.set_allowed_service(
                allowed_service, 'create')
            trusted_relationships['Statement'].append(trusted_services)
        if isinstance(trusted_relationships, dict):
            trusted_relationships = dumps(trusted_relationships)

        params = dict(RoleName=role_name,
                      AssumeRolePolicyDocument=trusted_relationships)
        if permissions_boundary:
            params['PermissionsBoundary'] = permissions_boundary

        try:
            role = self.client.create_role(**params)
            return role['Role']
        except ClientError as e:
            if e.response['Error']['Code'] == 'EntityAlreadyExists':
                return self.client.get_role(role_name)['Role']
            raise e

    @staticmethod
    def empty_trusted_relationships():
        trusted_relationships = {
            "Version": "2012-10-17",
            "Statement": []
        }
        return trusted_relationships

    def attach_policy(self, role_name, policy_arn):
        self.client.attach_role_policy(
            RoleName=role_name,
            PolicyArn=policy_arn
        )

    def attach_inline_policy(self, role_name, policy_name, policy_document):
        if isinstance(policy_document, dict):
            policy_document = dumps(policy_document)
        self.client.put_role_policy(
            RoleName=role_name,
            PolicyName=policy_name,
            PolicyDocument=policy_document
        )

    def detach_policy(self, role_name, policy_arn):
        self.client.detach_role_policy(
            RoleName=role_name,
            PolicyArn=policy_arn
        )

    def put_role_permissions_boundary(self, role_name, policy_arn):
        _LOG.info(f'Attaching permissions boundary policy: \'{policy_arn}\''
                  f' to role: \'{role_name}\'')
        self.client.put_role_permissions_boundary(
            RoleName=role_name,
            PermissionsBoundary=policy_arn
        )

    def delete_role_permissions_boundary(self, role_name):
        _LOG.info(f'Removing permissions boundary policy from \'{role_name}\'')
        try:
            self.client.delete_role_permissions_boundary(
                RoleName=role_name
            )
        except ClientError as e:
            if 'NoSuchEntity' in str(e):
                _LOG.warn(f'Role \'{role_name}\' doesn\'t have permissions '
                          f'boundary policy. Skipping...')
            else:
                _LOG.error(str(e))
                raise e

    def get_policy_arn(self, name, policy_scope='All'):
        """ Get policy arn from list existing. To reduce list result there is
        an ability to define policy scope.

        :type policy_scope: str
        :param policy_scope: 'All'|'AWS'|'Local'
        :type name: str
        """
        # TODO this method is highly time-ineffective especially if we, for
        #  instance, perform `syndicate transform` on a big meta, where
        #  there is a huge amount of policies names.
        #  lru_cache for self.get_policies makes the situation better but in
        #  general it should be refactored.
        custom_policies = self.get_policies(policy_scope)
        for each in custom_policies:
            if each['PolicyName'] == name:
                return each['Arn']

    def get_policy(self, arn):
        try:
            return self.client.get_policy(PolicyArn=arn)['Policy']
        except ClientError as e:
            if 'NoSuchEntity' in str(e):
                pass  # valid exception
            else:
                raise e

    def remove_policy_version(self, policy_arn, version_id):
        self.client.delete_policy_version(
            PolicyArn=policy_arn,
            VersionId=version_id
        )

    def create_policy_version(self, policy_arn, policy_document,
                              set_as_default=None):
        params = dict(PolicyArn=policy_arn,
                      PolicyDocument=policy_document)
        if set_as_default:
            params['SetAsDefault'] = set_as_default
        self.client.create_policy_version(**params)

    def remove_policy(self, policy_arn):
        """ To remove policy all it version must be removed before default one.

        :type policy_arn: str
        """
        try:
            version = self.client.list_policy_versions(PolicyArn=policy_arn)
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntity':
                _LOG.warn(f'Policy \'{policy_arn}\' not found')
                return
            else:
                raise e
        policy_versions = version['Versions']
        if policy_versions:
            for each in policy_versions:
                if each['IsDefaultVersion']:
                    continue
                self.remove_policy_version(policy_arn, each['VersionId'])
        self.client.delete_policy(PolicyArn=policy_arn)

    def remove_role(self, role_name):
        self.client.delete_role(RoleName=role_name)

    def create_instance_profile(self, profile_name):
        self.client.create_instance_profile(
            InstanceProfileName=profile_name
        )

    def remove_instance_profile(self, profile_name):
        self.client.delete_instance_profile(
            InstanceProfileName=profile_name
        )

    def get_instance_profiles(self):
        profiles = []
        response = self.client.list_instance_profiles()
        token = response.get('Marker')
        profiles.extend(response.get('InstanceProfiles'))
        while token:
            response = self.client.list_instance_profiles(Marker=token)
            token = response.get('Marker')
            profiles.extend(response.get('InstanceProfiles'))
        return profiles

    def is_instance_profile_exists(self, profile_name):
        profiles = self.get_instance_profiles()
        for each in profiles:
            if each['InstanceProfileName'] == profile_name:
                return each

    def add_role_to_instance_profile(self, profile_name, role_name):
        self.client.add_role_to_instance_profile(
            InstanceProfileName=profile_name,
            RoleName=role_name
        )

    def remove_role_from_instance_profile(self, profile_name, role_name):
        self.client.remove_role_from_instance_profile(
            InstanceProfileName=profile_name,
            RoleName=role_name
        )

    def get_instance_profiles_for_role(self, role_name):
        profiles = []
        response = self.client.list_instance_profiles_for_role(
            RoleName=role_name
        )
        token = response.get('Marker')
        profiles.extend(response.get('InstanceProfiles'))
        while token:
            response = self.client.list_instance_profiles_for_role(
                RoleName=role_name, Marker=token
            )
            token = response.get('Marker')
            profiles.extend(response.get('InstanceProfiles'))
        return profiles

    def get_assume_role_policy_document(self, role_name):
        return self.resource.Role(role_name).assume_role_policy_document

    def update_assume_role_policy_document(self, role_name, document):
        self.resource.AssumeRolePolicy(role_name).update(
            PolicyDocument=document)

    def create_user(self, name, path=None):
        params = dict(UserName=name)
        if path:
            params['Path'] = path
        return self.client.create_user(**params)

    def delete_user(self, name):
        self.client.delete_user(UserName=name)

    def attach_policy_to_user(self, name, policy_arn):
        params = dict(UserName=name, PolicyArn=policy_arn)
        self.client.attach_user_policy(**params)

    def create_access_key(self, user_name):
        return self.client.create_access_key(UserName=user_name)

    def get_users(self, path=None):
        users = []
        params = dict()
        if path:
            params['PathPrefix'] = path
        response = self.client.list_users(**params)
        users.extend(response.get('Users'))
        token = response.get('Marker')
        while token:
            params['Marker'] = token
            response = self.client.list_users(**params)
            token = response.get('Marker')
            users.extend(response.get('Users'))
        return users

    def is_user_exists(self, user_name):
        list_users = self.get_users()
        if list_users:
            for each in list_users:
                if user_name == each['UserName']:
                    return each['Arn']

    def get_access_keys(self, user_name):
        keys = []
        response = self.client.list_access_keys(UserName=user_name)
        keys.extend(response.get('AccessKeyMetadata'))
        token = response.get('Marker')
        while token:
            response = self.client.list_access_keys(UserName=user_name,
                                                    Marker=token)
            token = response.get('Marker')
            keys.extend(response.get('AccessKeyMetadata'))
        return keys

    def delete_access_key(self, user_name, access_key):
        self.client.delete_access_key(AccessKeyId=access_key,
                                      UserName=user_name)

    def get_user_attached_policies(self, user_name, path=None):
        policies = []
        params = dict(UserName=user_name)
        if path:
            params['PathPrefix'] = path
        response = self.client.list_attached_user_policies(**params)
        policies.extend(response.get('AttachedPolicies'))
        token = response.get('Marker')
        while token:
            params['Marker'] = token
            response = self.client.list_attached_user_policies(**params)
            token = response.get('Marker')
            policies.extend(response.get('AttachedPolicies'))
        return policies

    def detach_user_policy(self, user_name, policy_arn):
        self.client.detach_user_policy(UserName=user_name,
                                       PolicyArn=policy_arn)

    def get_user_certificates(self, user_name):
        certs = []
        response = self.client.list_signing_certificates(UserName=user_name)
        certs.extend(response.get('Certificates'))
        token = response.get('Marker')
        while token:
            response = self.client.list_signing_certificates(
                UserName=user_name,
                Marker=token)
            token = response.get('Marker')
            certs.extend(response.get('Certificates'))
        return certs

    def delete_user_certificate(self, user_name, cert_id):
        self.client.delete_signing_certificate(UserName=user_name,
                                               CertificateId=cert_id)

    def get_user_ssh_keys(self, user_name):
        ssh_keys = []
        response = self.client.list_ssh_public_keys(UserName=user_name)
        ssh_keys.extend(response.get('SSHPublicKeys'))
        token = response.get('Marker')
        while token:
            response = self.client.list_ssh_public_keys(UserName=user_name,
                                                        Marker=token)
            token = response.get('Marker')
            ssh_keys.extend(response.get('SSHPublicKeys'))
        return ssh_keys

    def delete_user_ssh_key(self, user_name, ssh_id):
        self.client.delete_ssh_public_key(UserName=user_name,
                                          SSHPublicKeyId=ssh_id)

    def get_user_inline_policies(self, user_name):
        policies = []
        response = self.client.list_user_policies(UserName=user_name)
        policies.extend(response.get('PolicyNames'))
        token = response.get('Marker')
        while token:
            response = self.client.list_user_policies(UserName=user_name,
                                                      Marker=token)
            token = response.get('Marker')
            policies.extend(response.get('PolicyNames'))
        return policies

    def update_custom_role(self, role, role_name, allowed_account=None,
                           allowed_service=None, trusted_relationships=None,
                           external_id=None):
        updated_role = role['AssumeRolePolicyDocument']
        if trusted_relationships:
            trusted_relationships = {
                "Version": "2012-10-17",
                "Statement": updated_role.get('Statement', [])
            }
        else:
            trusted_relationships = IAMConnection.empty_trusted_relationships()
        statement = trusted_relationships['Statement']
        if allowed_account:
            trusted_accounts = IAMConnection.set_allowed_account(
                allowed_account, external_id, 'update')
            statement.append(trusted_accounts)
        if allowed_service:
            trusted_services = IAMConnection.set_allowed_service(
                allowed_service, 'update')
            statement.append(trusted_services)
        if isinstance(trusted_relationships, dict):
            trusted_relationships = dumps(trusted_relationships)
            statement = updated_role.get('Statement', [])
            statement.append(trusted_relationships)
            unique = []
            for s in statement:
                if s not in unique:
                    unique.append(s)
        try:
            role = self.client.update_assume_role_policy(
                RoleName=role_name,
                PolicyDocument=trusted_relationships)
            return role['ResponseMetadata']
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntityException':
                _LOG.warn(f'Can not update role \'{role_name}\': role does '
                          f'not exist.')
            raise e

    @staticmethod
    def set_allowed_account(allowed_account, external_id, action):
        if isinstance(allowed_account, str):
            principal = get_account_role_arn(allowed_account)
        elif isinstance(allowed_account, list):
            principal = []
            for each in allowed_account:
                principal.append(get_account_role_arn(each))
        else:
            raise TypeError(
                f'Can not {action} role. \'allowed_account\' must be list '
                f'or string. Actual type: {type(allowed_account)}')
        trusted_accounts = {
            "Sid": "",
            "Effect": "Allow",
            "Principal": {
                "AWS": principal
            },
            "Action": "sts:AssumeRole"
        }
        if external_id:
            trusted_accounts['Condition'] = {
                "StringEquals": {
                    "sts:ExternalId": external_id
                }
            }
        return trusted_accounts

    @staticmethod
    def set_allowed_service(allowed_service, action):
        if isinstance(allowed_service, str):
            principal = "{0}.amazonaws.com".format(allowed_service)
        elif isinstance(allowed_service, list):
            principal = []
            for each in allowed_service:
                principal.append("{0}.amazonaws.com".format(each))
        else:
            raise TypeError(
                f'Can not {action} role. \'allowed_service\' must be list '
                f'or string. Actual type: {type(allowed_service)}')
        trusted_services = {
            "Effect": "Allow",
            "Principal": {
                "Service": principal
            },
            "Action": "sts:AssumeRole"
        }
        return trusted_services

    def update_custom_policy_content(self, name, arn, content):
        policy_resource = self.resource.Policy(arn)
        policy_json = policy_resource.default_version.document
        statement = content.get('Statement')
        if not statement:
            _LOG.warn(f'Policy \'{name}\' has no or empty \'Statement\' '
                      f'field.')
            statement = []
        policy_json['Statement'] = statement
        version = content.get('Version')
        if not statement:
            _LOG.warn(f'Policy \'{name}\' has no or empty \'Version\' '
                      f'field.')
            version = '2012-10-17'
        policy_json['Version'] = version
        policy = self.get_policy(arn=arn)
        policy_version = self.client.get_policy_version(
            PolicyArn=arn, VersionId=policy['DefaultVersionId']
        )['PolicyVersion']
        if content == policy_version['Document']:
            _LOG.warn(f'No need to update policy \'{name}\': the new and the '
                      f'old contents are identical.')
            return
        versions = self.client.list_policy_versions(PolicyArn=arn)["Versions"]
        to_remove = next((v for v in reversed(versions) if v['IsDefaultVersion'] == False), None)
        if to_remove:
            _LOG.info(f'Old version of policy is found. Removing one: {to_remove}')
            self.remove_policy_version(policy_arn=arn,
                                       version_id=to_remove['VersionId'])
        self.create_policy_version(policy_arn=arn,
                                   policy_document=dumps(policy_json),
                                   set_as_default=True)

    def create_group(self, name):
        return self.client.create_group(GroupName=name)

    def get_group(self, name):
        groups = []
        try:
            response = self.client.get_group(GroupName=name)
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntityException':
                _LOG.warn(f'Group {name} is not found')
                return []
            raise e
        token = response.get('Marker')
        group_item = response.get('Group')
        group_item.update({'Users': response.get('Users')})
        groups.extend(group_item)
        while token:
            response = self.client.get_group(GroupName=name, Marker=token)
            token = response.get('Marker')
            group_item = response.get('Group')
            group_item.update({'Users': response.get('Users')})
            groups.extend(group_item)
        return groups

    def add_user_to_group(self, group_name, username):
        try:
            response = self.client.add_user_to_group(GroupName=group_name,
                                                     UserName=username)
            return response
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntityException':
                _LOG.warn(f'Group {group_name} or username {username} is not '
                          f'found')
                return []
            raise e

    def remove_user_from_group(self, group_name, username):
        try:
            response = self.client.remove_user_from_group(GroupName=group_name,
                                                          UserName=username)
            return response
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntityException':
                _LOG.warn(f'Group {group_name} or username {username} is not '
                          f'found')
            raise e

    def attach_group_policy(self, group_name, arn):
        try:
            response = self.client.attach_group_policy(GroupName=group_name,
                                                       PolicyArn=arn)
            return response
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntityException':
                _LOG.warn(f'Group {group_name} is not found')
            elif e.response['Error']['Code'] == 'LimitExceededException':
                _LOG.warn(f'Can not attach more than 10 rules to group '
                          f'{group_name}')
            raise e

    def get_waiter(self, waiter_name):
        return self.client.get_waiter(waiter_name)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.conf.validator import ALL_REGIONS
from syndicate.core.resources.helper import check_region_available


def deploy_regions(resource_name, meta):
    from syndicate.core import CONFIG
    regions = []
    region = meta.get('region')
    if region is None:
        regions.append(CONFIG.region)
    elif isinstance(region, str):
        if region == 'all':
            for each in ALL_REGIONS:
                regions.append(each)
        else:
            if check_region_available(region, ALL_REGIONS, meta):
                regions.append(region)
    elif isinstance(region, list):
        for each in region:
            if check_region_available(each, ALL_REGIONS, meta):
                regions.append(each)
    else:
        raise AssertionError(
            'Invalid value region: {0}. Resource: {1}.'.format(region,
                                                               resource_name))
    return regions



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.commons.log_helper import get_logger
from syndicate.core.resources.batch_compenv_resource import DEFAULT_STATE
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_resource_name_builder import \
    build_terraform_resource_name
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_role_name_ref, build_instance_profile_arn_ref, build_role_arn_ref

_LOG = get_logger('syndicate.core.transform.terraform'
                  '.converter.tf_batch_compenv_converter')

ECS_POLICY_ARN = 'arn:aws:iam::aws:policy/service-role/AmazonEC2ContainerServiceforEC2Role'
BATCH_SERVICE_ROLE = 'arn:aws:iam::aws:policy/service-role/AWSBatchServiceRole'
AWS_BATCH_SERVICE_ROLE = 'AWSBatchServiceRole'


class BatchComputeEnvConverter(TerraformResourceConverter):

    def convert(self, name, resource):
        state = resource.get('state')
        if not state:
            state = DEFAULT_STATE

        service_role_arn = self.service_role_arn(resource_meta=resource)

        env_type = resource.get('compute_environment_type')
        compute_resources = resource.get('compute_resources', {})
        res_type = compute_resources.get('type')
        minv_cpus = compute_resources.get('minv_cpus')
        desired_vcpus = compute_resources.get('desired_vcpus')
        maxv_cpus = compute_resources.get('maxv_cpus')
        instance_types = compute_resources.get('instance_types', [])
        security_group_ids = compute_resources.get('security_group_ids', [])
        subnets = compute_resources.get('subnets', [])

        instance_role = compute_resources.get('instance_role')
        if instance_role:
            ecs_policy_attachment = aws_iam_role_policy_attachment(
                role_name=instance_role, policy_arn=ECS_POLICY_ARN)
            self.template.add_aws_iam_role_policy_attachment(
                meta=ecs_policy_attachment)

        profile_name = f'{instance_role}_instance_profile'
        profile = aws_iam_instance_profile(role_name=instance_role,
                                           profile_name=profile_name)
        self.template.add_aws_iam_instance_profile(meta=profile)

        aws_batch_compute_environment = batch_com_env(
            compute_environment_name=name, env_type=env_type, state=state,
            service_role=service_role_arn,
            res_type=res_type,
            maxv_cpus=maxv_cpus,
            desired_vcpus=desired_vcpus, minv_cpus=minv_cpus,
            instance_types=instance_types,
            security_group_ids=security_group_ids, subnets=subnets,
            instance_profile_name=profile_name)

        self.template.add_aws_batch_compute_environment(
            meta=aws_batch_compute_environment)

    def service_role_arn(self, resource_meta):
        service_role = resource_meta.get('service_role')
        if not service_role:
            role = self.template.get_resource_by_name(AWS_BATCH_SERVICE_ROLE)
            if not role:
                batch_service_role = aws_batch_service_role(
                    AWS_BATCH_SERVICE_ROLE)
                self.template.add_aws_iam_role(meta=batch_service_role)
                batch_service_role_policy_attachment = aws_iam_role_policy_attachment(
                    role_name=AWS_BATCH_SERVICE_ROLE,
                    policy_arn=BATCH_SERVICE_ROLE)
                self.template.add_aws_iam_role_policy_attachment(
                    batch_service_role_policy_attachment)
                return build_role_arn_ref(AWS_BATCH_SERVICE_ROLE)
        role = self.template.get_resource_by_name(service_role)
        if not role:
            raise AssertionError("IAM role '{}' is not present "
                                 "in build meta.".format(service_role))
        return build_role_arn_ref(role_name=role)


def aws_batch_service_role(role_name):
    policy_content = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Action": "sts:AssumeRole",
                "Effect": "Allow",
                "Principal": {
                    "Service": "batch.amazonaws.com"
                }
            }
        ]
    }
    resource = {
        role_name: {
            "name": role_name,
            "assume_role_policy": json.dumps(policy_content)
        }
    }
    return resource


def aws_iam_role_policy_attachment(role_name, policy_arn):
    resource = {
        build_terraform_resource_name(role_name, 'policy_attachment'):
            {
                "policy_arn": policy_arn,
                "role": build_role_name_ref(role_name)
            }
    }
    return resource


def aws_iam_instance_profile(profile_name, role_name):
    resource = {
        profile_name: {
            "name": role_name,
            "role": build_role_name_ref(role_name)
        }
    }
    return resource


def batch_com_env(compute_environment_name, env_type, service_role, state,
                  res_type=None, minv_cpus=None, desired_vcpus=None,
                  maxv_cpus=None,
                  instance_types=None, security_group_ids=None, subnets=None,
                  instance_profile_name=None):
    params = {
        'compute_environment_name': compute_environment_name,
        'type': env_type,
        'state': state,
        'service_role': service_role
    }

    comp_resources = dict()
    if res_type:
        comp_resources['type'] = res_type
    if minv_cpus:
        comp_resources['max_vcpus'] = minv_cpus
    if maxv_cpus:
        comp_resources['max_vcpus'] = maxv_cpus
    if desired_vcpus:
        comp_resources['desired_vcpus'] = desired_vcpus
    if instance_types:
        comp_resources['instance_type'] = instance_types
    if security_group_ids:
        comp_resources['security_group_ids'] = security_group_ids
    if subnets:
        comp_resources['subnets'] = subnets
    if instance_profile_name:
        comp_resources['instance_role'] = build_instance_profile_arn_ref(
            instance_profile_name=instance_profile_name)

    if comp_resources:
        params['compute_resources'] = comp_resources

    resource = {
        compute_environment_name: params
    }
    return resource



from syndicate.commons.log_helper import get_logger
from syndicate.core.generators.contents import S3_BUCKET_PUBLIC_READ_POLICY
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import S3_BUCKET_TYPE

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.base_generator')


class S3Generator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = S3_BUCKET_TYPE
    CONFIGURATION = {
        'acl': 'private',
        'location': None,
        'cors': list,
        'policy': dict,
        'public_access_block': {
            'block_public_acls': True,
            'ignore_public_acls': True,
            'block_public_policy': True,
            'restrict_public_buckets': True
        }
    }

    def __init__(self, **kwargs):
        self.static_website_hosting = kwargs.get('static_website_hosting')
        super().__init__(**kwargs)

    def _generate_resource_configuration(self) -> dict:
        result = super()._generate_resource_configuration()
        if self.static_website_hosting:
            result['policy'] = S3_BUCKET_PUBLIC_READ_POLICY
            result['acl'] = 'public-read'
            result['public_access_block'] = {
                'block_public_acls': False,
                'ignore_public_acls': False,
                'block_public_policy': False,
                'restrict_public_buckets': False
            }
            result['website_hosting'] = {
                'enabled': True,
                'index_document': 'index.html',
                'error_document': 'error.html'
            }
            _LOG.info(f'Deployment resources of the S3 bucket '
                      f'\'{self.resource_name}\' modified for static website '
                      f'hosting.')
        return result



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.documentdb_resource')


class DocumentDBInstanceResource(BaseResource):

    def __init__(self, docdb_conn, account_id,
                 region) -> None:
        self.connection = docdb_conn
        self.account_id = account_id
        self.region = region

    def create_db_instance(self, args):
        """ Create DocumentDB instance in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_db_instance_from_meta,
                                args)

    def describe_documentdb_instance(self, identifier, meta):
        if not identifier:
            return
        response = self.connection.describe_db_instances(identifier)
        if not response:
            return
        arn = f'arn:aws:rds:{self.region}:{self.account_id}:' \
              f'db:{identifier}'
        return {
            arn: build_description_obj(response, identifier, meta)
        }

    @unpack_kwargs
    def _create_db_instance_from_meta(self, name, meta):
        instance = self.connection.describe_db_instances(name)
        if instance:
            _LOG.warn(f'\'{name}\' instance exists.')
            return self.describe_documentdb_instance(identifier=name,
                                                     meta=meta)

        _LOG.info(f'Creating documentDB instance {name}')
        cluster_identifier = meta.get('cluster_identifier', None)
        instance_class = meta.get('instance_class', None)
        availability_zone = meta.get('availability_zone', None)
        instance = self.connection.create_db_instance(
            instance_identifier=name, availability_zone=availability_zone,
            instance_class=instance_class,
            cluster_identifier=cluster_identifier)

        _LOG.info(f'Created documentDB instance {instance}')
        return self.describe_documentdb_instance(identifier=name, meta=meta)

    def remove_db_instance(self, args):
        self.create_pool(self._remove_db_instance, args)

    @unpack_kwargs
    def _remove_db_instance(self, arn, config):
        instance = config['description']
        try:
            self.connection.delete_db_instance(instance)
            _LOG.info(f'DocumentDB instance \'{instance}\' was removed')
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn(f'DocumentDB instance \'{instance}\' is not found')
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os
from pathlib import Path

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core import ProjectState
from syndicate.core.project_state.project_state import BUILD_MAPPINGS
from syndicate.core.generators import (_touch,
                                       _mkdir, _write_content_to_file,
                                       FILE_LAMBDA_HANDLER_PYTHON,
                                       FILE_LAMBDA_HANDLER_NODEJS,
                                       _read_content_from_file)
from syndicate.core.generators.project import _generate_java_project_hierarchy
from syndicate.core.generators.tests import _generate_python_tests
from syndicate.core.generators.contents import (
    NODEJS_LAMBDA_HANDLER_TEMPLATE,
    _generate_python_node_lambda_config,
    _generate_lambda_role_config, _generate_nodejs_node_lambda_config,
    CANCEL_MESSAGE, _generate_package_nodejs_lambda,
    _generate_package_lock_nodejs_lambda, JAVA_LAMBDA_HANDLER_CLASS,
    SRC_MAIN_JAVA, FILE_POM, PYTHON_LAMBDA_HANDLER_TEMPLATE, INIT_CONTENT,
    ABSTRACT_LAMBDA_CONTENT, EXCEPTION_CONTENT, LOG_HELPER_CONTENT,
    _generate_python_node_layer_config, REQUIREMENTS_FILE_CONTENT,
    LOCAL_REQUIREMENTS_FILE_CONTENT, _generate_node_layer_package_file,
    _generate_node_layer_package_lock_file)
from syndicate.core.groups import (RUNTIME_JAVA, RUNTIME_NODEJS,
                                   RUNTIME_PYTHON, RUNTIME_PYTHON_LAYER,
                                   RUNTIME_NODEJS_LAYER)

_LOG = get_logger('syndicate.core.generators.lambda_function')
USER_LOG = get_user_logger()

FOLDER_LAMBDAS = 'lambdas'
FOLDER_COMMONS = 'commons'
FOLDER_LAYERS = 'layers'

FILE_DEPLOYMENT_RESOURCES = 'deployment_resources.json'
FILE_INIT_PYTHON = '__init__.py'
FILE_LAMBDA_CONFIG = 'lambda_config.json'
FILE_PACKAGE_LOCK = 'package-lock.json'
FILE_PACKAGE = 'package.json'
FILE_LAYER_CONFIG = 'lambda_layer_config.json'

FILE_REQUIREMENTS = 'requirements.txt'
FILE_LOCAL_REQUIREMENTS = 'local_requirements.txt'

LAMBDA_ROLE_NAME_PATTERN = '{0}-role'  # 0 - lambda_name
POLICY_NAME_PATTERN = '{0}-policy'  # 0 - lambda_name

ABSTRACT_LAMBDA_NAME = 'abstract_lambda'

PROJECT_STATE_PARAM = 'project_state'
LAMBDAS_PATH_PARAM = 'lambdas_path'
LAMBDA_NAMES_PARAM = 'lambda_names'
PROJECT_NAME_PARAM = 'project_name'
PROJECT_PATH_PARAM = 'project_path'
LAYER_NAME_PARAM = 'layer_name'
LAYERS_PATH_PARAM = 'layers_path'
RUNTIME_PARAM = 'runtime'

PYTHON_LAMBDA_FILES = [
    FILE_INIT_PYTHON, FILE_LOCAL_REQUIREMENTS,
    FILE_REQUIREMENTS,
    FILE_DEPLOYMENT_RESOURCES,  # content
    FILE_LAMBDA_CONFIG  # content
]  # + handler content

PYTHON_LAYER_FILES = [
    FILE_LOCAL_REQUIREMENTS,
    FILE_REQUIREMENTS,
    FILE_LAYER_CONFIG
]

NODEJS_LAMBDA_FILES = [
    FILE_PACKAGE,
    FILE_PACKAGE_LOCK,
    FILE_LAMBDA_CONFIG,
    FILE_LAMBDA_HANDLER_NODEJS,
    FILE_DEPLOYMENT_RESOURCES
]

NODEJS_LAYER_FILES = [
    FILE_PACKAGE,
    FILE_PACKAGE_LOCK,
    FILE_LAYER_CONFIG
]


def generate_common_module(src_path, runtime):
    runtime_processor = COMMON_MODULE_PROCESSORS.get(runtime)
    if not runtime_processor:
        raise AssertionError(f'Unable to create a common module in {src_path}')
    runtime_processor(src_path=src_path)


def generate_lambda_function(project_path, runtime,
                             lambda_names):
    if not os.path.exists(project_path):
        USER_LOG.info(f'Project "{project_path}" you '
                      f'have provided does not exist')
        return

    if not ProjectState.check_if_project_state_exists(
            project_path=project_path):
        USER_LOG.info(f'Seems that the path {project_path} is not a project')
        return
    project_state = ProjectState(project_path=project_path)
    src_path = os.path.join(project_path, BUILD_MAPPINGS[runtime])

    common_module_generator = COMMON_MODULE_PROCESSORS.get(runtime)
    if not common_module_generator:
        raise AssertionError(f'The runtime {runtime} is not currently '
                             f'supported to bootstrap the project')
    common_module_generator(src_path=src_path)
    project_state.add_project_build_mapping(runtime=runtime)

    processor = LAMBDAS_PROCESSORS.get(runtime)
    if not processor:
        raise RuntimeError(f'Wrong project runtime {runtime}')

    lambdas_path = os.path.join(src_path, FOLDER_LAMBDAS)

    processor(project_path=project_path, lambda_names=lambda_names,
              lambdas_path=lambdas_path, project_state=project_state)

    tests_generator = TESTS_MODULE_PROCESSORS.get(runtime)
    [tests_generator(project_path, name) for name in lambda_names]

    project_state.save()
    if len(lambda_names) == 1:
        USER_LOG.info(f'Lambda {lambda_names[0]} has been successfully '
                      f'added to the project.')
    else:
        generated_lambdas = ', '.join(lambda_names)
        USER_LOG.info(f'The following lambdas have been successfully '
                      f'added to the project: {generated_lambdas}')


def generate_lambda_layer(name, runtime, project_path, lambda_names=None):
    if not os.path.exists(project_path):
        USER_LOG.info(f'Project "{project_path}" you '
                      f'have provided does not exist')
        return

    if not ProjectState.check_if_project_state_exists(
            project_path=project_path):
        USER_LOG.info(f'Seems that the path {project_path} is not a project')
        return
    project_state = ProjectState(project_path=project_path)
    src_path = os.path.join(project_path, BUILD_MAPPINGS[runtime])

    common_module_generator = COMMON_MODULE_PROCESSORS.get(runtime + '_layer')
    if not common_module_generator:
        raise AssertionError(f'The layer runtime {runtime} is not currently '
                             f'supported to bootstrap the project')
    common_module_generator(src_path=src_path)
    project_state.add_project_build_mapping(runtime=runtime)

    processor = LAYERS_PROCESSORS.get(runtime)
    if not processor:
        raise RuntimeError(f'Wrong layer runtime {runtime}')

    layers_path = os.path.join(src_path, FOLDER_LAMBDAS, FOLDER_LAYERS)

    processor(layer_name=name, layers_path=layers_path, runtime=runtime)

    project_state.save()

    USER_LOG.info(f'Layer \'{name}\' has been successfully '
                  f'added to the project.')

    if lambda_names:
        _LOG.debug(f'Going to link layer {name} with lambdas: {lambda_names}')
        project_lambdas = project_state.lambdas
        _link_layer_to_lambdas(lambda_names=lambda_names,
                               layer_name=name,
                               layer_runtime=runtime,
                               existent_lambdas=project_lambdas,
                               lambda_path=src_path)


def _generate_python_lambdas(**kwargs):
    lambdas_path = kwargs.get(LAMBDAS_PATH_PARAM)
    lambda_names = kwargs.get(LAMBDA_NAMES_PARAM)
    project_state = kwargs.get(PROJECT_STATE_PARAM)

    if not os.path.exists(lambdas_path):
        _mkdir(lambdas_path, exist_ok=True)

    init_file = os.path.join(lambdas_path, '__init__.py')
    _touch(init_file)
    _LOG.info(f'Lambdas path: {lambdas_path}')
    for lambda_name in lambda_names:
        lambda_folder = os.path.join(lambdas_path, lambda_name)

        answer = _mkdir(
            path=lambda_folder,
            fault_message=f'\nLambda {lambda_name} already exists.\nOverride '
                          f'the Lambda function? [y/n]: ')
        if not answer:
            _LOG.info(CANCEL_MESSAGE.format(lambda_name))
            continue

        PYTHON_LAMBDA_FILES.append(
            FILE_LAMBDA_HANDLER_PYTHON)  # add lambda handler
        for file in PYTHON_LAMBDA_FILES:
            _touch(os.path.join(lambda_folder, file))

        # fill handler.py
        lambda_class_name = __lambda_name_to_class_name(
            lambda_name=lambda_name)
        python_lambda_handler_template = PYTHON_LAMBDA_HANDLER_TEMPLATE. \
            replace('LambdaName', lambda_class_name)
        _write_content_to_file(os.path.join(
            lambda_folder, FILE_LAMBDA_HANDLER_PYTHON),
            python_lambda_handler_template)

        # fill deployment_resources.json
        pattern_format = LAMBDA_ROLE_NAME_PATTERN.format(lambda_name)
        role_def = _generate_lambda_role_config(pattern_format)
        _write_content_to_file(os.path.join(
            lambda_folder, FILE_DEPLOYMENT_RESOURCES), role_def)

        # fill lambda_config.json
        lambda_def = _generate_python_node_lambda_config(
            lambda_name,
            os.path.join(FOLDER_LAMBDAS, lambda_name))
        _write_content_to_file(os.path.join(lambda_folder, FILE_LAMBDA_CONFIG),
                               lambda_def)

        # fill local_dependencies.txt
        _write_content_to_file(
            os.path.join(lambda_folder, FILE_LOCAL_REQUIREMENTS),
            FOLDER_COMMONS)

        project_state.add_lambda(lambda_name=lambda_name, runtime='python')

        _LOG.info(f'Lambda {lambda_name} created')


def __lambda_name_to_class_name(lambda_name):
    class_name = lambda_name.split('-')
    if len(class_name) == 1:
        class_name = lambda_name.split('_')
    return ''.join([_.capitalize() for _ in class_name])


def _generate_java_lambdas(**kwargs):
    project_path = kwargs.get(PROJECT_PATH_PARAM)
    project_state = kwargs.get(PROJECT_STATE_PARAM)
    project_name = project_state.name
    lambda_names = kwargs.get(LAMBDA_NAMES_PARAM, [])

    _generate_java_project_hierarchy(project_name=project_name,
                                     full_project_path=project_path)

    java_package_name = _generate_java_package_name(project_name)
    java_package_as_path = java_package_name.replace('.', '/')

    pom_file_path = os.path.join(project_path, FILE_POM)
    pom_xml_content = _read_content_from_file(pom_file_path)
    pom_xml_content = pom_xml_content.replace(
        '<!--packages to scan-->',
        f'<package>{java_package_name}</package>')
    _write_content_to_file(pom_file_path, pom_xml_content)

    full_package_path = Path(project_path, SRC_MAIN_JAVA, java_package_as_path)
    for lambda_name in lambda_names:
        if not os.path.exists(full_package_path):
            _mkdir(full_package_path, exist_ok=True)

        lambda_class_name = _get_parts_split_by_chars(to_split=lambda_name,
                                                      chars=['-', '_']).title()
        lambda_class_name = lambda_class_name.replace(' ', '')
        lambda_role_name = LAMBDA_ROLE_NAME_PATTERN.format(lambda_name)
        java_handler_content = \
            (JAVA_LAMBDA_HANDLER_CLASS
             .replace('{java_package_name}', java_package_name)
             .replace('{lambda_name}', lambda_name)
             .replace('{lambda_class_name}', lambda_class_name)
             .replace('{lambda_role_name}', lambda_role_name))

        java_handler_file_name = os.path.join(
            project_path, SRC_MAIN_JAVA, java_package_as_path,
            f'{lambda_class_name}.java')
        _write_content_to_file(
            java_handler_file_name,
            java_handler_content
        )

        # add role to deployment_resource.json

        dep_res_path = os.path.join(project_path,
                                    FILE_DEPLOYMENT_RESOURCES)
        deployment_resources = json.loads(_read_content_from_file(
            dep_res_path
        ))
        deployment_resources.update(_generate_lambda_role_config(
            lambda_role_name, stringify=False))
        _write_content_to_file(dep_res_path,
                               json.dumps(deployment_resources, indent=2))

        project_state.add_lambda(lambda_name=lambda_name, runtime=RUNTIME_JAVA)
        _LOG.info(f'Lambda {lambda_name} created')


def _generate_java_package_name(project_name):
    unified_package_name = _get_parts_split_by_chars(to_split=project_name,
                                                     chars=['-', '_'])
    java_package_name = unified_package_name.replace(' ', '')
    java_package_name = f'com.{java_package_name}'
    return java_package_name


def _get_parts_split_by_chars(chars, to_split):
    result = to_split
    for char in chars:
        result = result.replace(char, ' ')
    return result


def _generate_nodejs_lambdas(**kwargs):
    lambdas_path = kwargs.get(LAMBDAS_PATH_PARAM)
    lambda_names = kwargs.get(LAMBDA_NAMES_PARAM, [])
    project_state = kwargs.get(PROJECT_STATE_PARAM)

    if not os.path.exists(lambdas_path):
        _mkdir(lambdas_path, exist_ok=True)
    for lambda_name in lambda_names:

        lambda_folder = os.path.join(lambdas_path, lambda_name)

        answer = _mkdir(
            path=lambda_folder,
            fault_message=f'\nLambda {lambda_name} already exists.\n'
                          f'Override the Lambda function? [y/n]: ')
        if not answer:
            _LOG.info(CANCEL_MESSAGE.format(lambda_name))
            continue

        for file in NODEJS_LAMBDA_FILES:
            _touch(Path(lambda_folder, file))

        # fill index.js
        _write_content_to_file(
            os.path.join(lambda_folder, FILE_LAMBDA_HANDLER_NODEJS),
            NODEJS_LAMBDA_HANDLER_TEMPLATE)

        # fill package.json
        package_def = _generate_package_nodejs_lambda(lambda_name)
        _write_content_to_file(os.path.join(lambda_folder, FILE_PACKAGE),
                               package_def)

        # fill package.json
        package_lock_def = _generate_package_lock_nodejs_lambda(
            lambda_name)
        _write_content_to_file(os.path.join(lambda_folder, FILE_PACKAGE_LOCK),
                               package_lock_def)

        # fill deployment_resources.json
        role_def = _generate_lambda_role_config(
            LAMBDA_ROLE_NAME_PATTERN.format(lambda_name))
        _write_content_to_file(
            os.path.join(lambda_folder, FILE_DEPLOYMENT_RESOURCES),
            role_def)

        # fill lambda_config.json
        lambda_def = _generate_nodejs_node_lambda_config(
            lambda_name,
            os.path.join(FOLDER_LAMBDAS, lambda_name))
        _write_content_to_file(os.path.join(lambda_folder, FILE_LAMBDA_CONFIG),
                               lambda_def)
        project_state.add_lambda(lambda_name=lambda_name, runtime=RUNTIME_NODEJS)
        _LOG.info(f'Lambda {lambda_name} created')


def _generate_python_layer(**kwargs):
    layer_name = kwargs.get(LAYER_NAME_PARAM)
    layers_path = kwargs.get(LAYERS_PATH_PARAM)
    runtime = kwargs.get(RUNTIME_PARAM)

    layer_folder = os.path.join(layers_path, layer_name)
    answer = _mkdir(
        path=layer_folder,
        fault_message=f'\nLayer \'{layer_name}\' already exists.\n'
                      f'Override? [y/n]: ')

    if not answer:
        USER_LOG.info(f'Creation of the layer \'{layer_name}\' skipped')
        return

    for file in PYTHON_LAYER_FILES:
        _touch(Path(layer_folder, file))

    layer_config = _generate_python_node_layer_config(layer_name, runtime)
    _write_content_to_file(os.path.join(layer_folder, FILE_LAYER_CONFIG),
                           layer_config)

    _write_content_to_file(os.path.join(layer_folder, FILE_REQUIREMENTS),
                           REQUIREMENTS_FILE_CONTENT)

    _write_content_to_file(os.path.join(layer_folder, FILE_LOCAL_REQUIREMENTS),
                           LOCAL_REQUIREMENTS_FILE_CONTENT)


def _generate_nodejs_layer(**kwargs):
    layer_name = kwargs.get(LAYER_NAME_PARAM)
    layers_path = kwargs.get(LAYERS_PATH_PARAM)
    runtime = kwargs.get(RUNTIME_PARAM)

    layer_folder = os.path.join(layers_path, layer_name)
    answer = _mkdir(
        path=layer_folder,
        fault_message=f'\nLayer \'{layer_name}\' already exists.\n'
                      f'Override? [y/n]: ')

    if not answer:
        USER_LOG.info(f'Creation of the layer \'{layer_name}\' skipped')
        return

    for file in NODEJS_LAYER_FILES:
        _touch(Path(layer_folder, file))

    layer_config = _generate_python_node_layer_config(layer_name, runtime)
    _write_content_to_file(os.path.join(layer_folder, FILE_LAYER_CONFIG),
                           layer_config)

    _write_content_to_file(os.path.join(layer_folder, FILE_PACKAGE),
                           _generate_node_layer_package_file(layer_name))

    _write_content_to_file(os.path.join(layer_folder, FILE_PACKAGE_LOCK),
                           _generate_node_layer_package_lock_file(layer_name))


def _link_layer_to_lambdas(lambda_names, layer_name, layer_runtime,
                           existent_lambdas, lambda_path):
    for lambda_name in lambda_names:
        if lambda_name not in existent_lambdas:
            USER_LOG.warn(f'The layer \'{layer_name}\' can\'t be linked with '
                          f'lambda function \'{lambda_name}\' due to an '
                          f'absence the function in the project.')
            continue

        lambda_runtime = existent_lambdas[lambda_name][RUNTIME_PARAM]
        if lambda_runtime != layer_runtime:
            USER_LOG.warn(f'The layer \'{layer_name}\' with runtime '
                          f'\'{layer_runtime}\' can\'t be linked with lambda '
                          f'\'{lambda_name}\' with runtime '
                          f'\'{lambda_runtime}\'')
            continue

        config_file_path = Path(lambda_path, FOLDER_LAMBDAS,
                                lambda_name, FILE_LAMBDA_CONFIG)
        if not os.path.isfile(config_file_path):
            USER_LOG.warn(f'The layer \'{layer_name}\' can\'t be linked with '
                          f'lambda function \'{lambda_name}\' due to an '
                          f'absence the function config file.')
            continue

        lambda_config = json.loads(_read_content_from_file(config_file_path))
        layers = lambda_config.get('layers')
        if isinstance(layers, list):
            layers.append(layer_name)
        else:
            layers = [layer_name]
        lambda_config['layers'] = list(set(layers))
        _write_content_to_file(config_file_path, json.dumps(lambda_config))
        USER_LOG.info(f'The layer \'{layer_name}\' was linked with '
                      f'\'{lambda_name}\'')


LAMBDAS_PROCESSORS = {
    RUNTIME_JAVA: _generate_java_lambdas,
    RUNTIME_NODEJS: _generate_nodejs_lambdas,
    RUNTIME_PYTHON: _generate_python_lambdas,
}

LAYERS_PROCESSORS = {
    RUNTIME_NODEJS: _generate_nodejs_layer,
    RUNTIME_PYTHON: _generate_python_layer
}


def _common_java_module(src_path):
    pass


def _common_nodejs_module(src_path):
    pass


def _common_python_module(src_path):
    common_module_path = os.path.join(src_path, FOLDER_COMMONS)
    _mkdir(path=common_module_path, exist_ok=True)

    init_path = os.path.join(common_module_path, '__init__.py')
    _touch(init_path)
    _write_content_to_file(file=init_path, content=INIT_CONTENT)

    abstract_lambda_path = os.path.join(common_module_path,
                                        'abstract_lambda.py')
    _touch(path=abstract_lambda_path)
    _write_content_to_file(file=abstract_lambda_path,
                           content=ABSTRACT_LAMBDA_CONTENT)

    logger_path = os.path.join(common_module_path, 'log_helper.py')
    _touch(path=logger_path)
    _write_content_to_file(file=logger_path, content=LOG_HELPER_CONTENT)

    exception_path = os.path.join(common_module_path, 'exception.py')
    _touch(path=exception_path)
    _write_content_to_file(file=exception_path, content=EXCEPTION_CONTENT)


def resolve_lambda_path(project: Path, runtime: str, source: str) -> Path:
    _lambda = ''
    if runtime == RUNTIME_JAVA:
        _lambda = _generate_java_package_name(project.name).replace('.', '/')
    elif runtime in LAMBDAS_PROCESSORS:
        _lambda = FOLDER_LAMBDAS
    return project/Path(source, _lambda)


def _common_python_nodejs_layer_module(src_path):
    layer_path = os.path.join(src_path, FOLDER_LAMBDAS, FOLDER_LAYERS)
    _mkdir(path=layer_path, exist_ok=True)


COMMON_MODULE_PROCESSORS = {
    RUNTIME_JAVA: _common_java_module,
    RUNTIME_NODEJS: _common_nodejs_module,
    RUNTIME_PYTHON: _common_python_module,
    RUNTIME_PYTHON_LAYER: _common_python_nodejs_layer_module,
    RUNTIME_NODEJS_LAYER: _common_python_nodejs_layer_module

}

TESTS_MODULE_PROCESSORS = {
    RUNTIME_JAVA: lambda project_path, lambda_name: None,
    RUNTIME_NODEJS: lambda project_path, lambda_name: None,
    RUNTIME_PYTHON: _generate_python_tests,
}



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import CLOUD_WATCH_RULE_TYPE
import click

DEFAULT_CRON_EXPRESSION = 'rate(1 hour)'


class CloudwatchEventRuleGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = CLOUD_WATCH_RULE_TYPE
    CONFIGURATION = {
        "rule_type": None,
        "region": None
    }

    def __init__(self, **kwargs):
        rule_type = kwargs.get('rule_type')
        if rule_type == 'schedule':
            self.CONFIGURATION['expression'] = DEFAULT_CRON_EXPRESSION
        elif rule_type == 'ec2':
            self.CONFIGURATION['instance_ids'] = list
            self.CONFIGURATION['instance_states'] = list
        elif rule_type == 'api_call':
            if not kwargs.get('aws_service'):
                raise click.MissingParameter(
                    "'aws_service' option is required "
                    "if if rule_type is 'api_call'",
                    param_type='option',
                    param_hint='aws_service')
            self.CONFIGURATION['aws_service'] = None
            self.CONFIGURATION['operations'] = list
        super().__init__(**kwargs)



import click
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import DOCUMENTDB_INSTANCE_TYPE, \
    DOCUMENTDB_CLUSTER_TYPE
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.docdb_instance_generator')
USER_LOG = get_user_logger()


class DocumentDBInstanceGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = DOCUMENTDB_INSTANCE_TYPE
    CONFIGURATION = {
        "cluster_identifier": None,
        "instance_class": "db.r5.large",
        "availability_zone": None
    }

    def write(self):
        cluster_identifier = self._dict.get('cluster_identifier')
        paths_with_cluster = self._get_resource_meta_paths(
            cluster_identifier,
            DOCUMENTDB_CLUSTER_TYPE)
        if not paths_with_cluster:
            message = f"Cluster '{cluster_identifier}' hasn't been declared " \
                      f"in deployment resources yet."
            _LOG.warn(message)
            if click.confirm(f"{message} Write DocumentDB instance anyway?"):
                USER_LOG.warn(f"Writing instance '{self.resource_name}' "
                              f"despite not declared cluster "
                              f"'{cluster_identifier}'")
            else:
                USER_LOG.warn(f"Skipping instance '{self.resource_name}'...")
                raise RuntimeError
        super().write()



from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs, delete_none
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.firehose_resource')


class FirehoseResource(BaseResource):

    def __init__(self, firehose_conn, s3_resource, iam_resource) -> None:
        self.connection = firehose_conn
        self.s3_resource = s3_resource
        self.iam_resource = iam_resource

    def create_stream(self, args):
        return self.create_pool(self._create_stream_from_meta, args)

    @unpack_kwargs
    def _create_stream_from_meta(self, name, meta):
        response = self.connection.describe_delivery_stream(name)
        if response:
            _arn = response['DeliveryStreamARN']
            return self.describe_stream(name, meta, _arn, response)

        _s3 = meta.get('s3_destination_configuration') or {}
        s3_configuration = delete_none({
            'RoleARN': self.iam_resource.build_role_arn(_s3.get('role')),
            'BucketARN': self.s3_resource.build_bucket_arn(_s3.get('bucket')),
            'Prefix': _s3.get('prefix'),
            'ErrorOutputPrefix': _s3.get('error_output_prefix'),
            'BufferingHints': {
                'SizeInMBs': (_s3.get('buffering_hints') or {}).get('size_in_mbs') or 5,
                'IntervalInSeconds': (_s3.get('buffering_hints') or {}).get('interval_in_seconds') or 300
            },
            'CompressionFormat': _s3.get('compression_format') or 'UNCOMPRESSED',
            # 'EncryptionConfiguration': {},
            # 'CloudWatchLoggingOptions': {}
        })
        _kinesis = meta.get('kinesis_stream_source_configuration') or {}
        kinesis_configuration = delete_none({
            'KinesisStreamARN': _kinesis.get('kinesis_stream_arn'),
            'RoleARN': self.iam_resource.build_role_arn(_kinesis.get('role'))
        })
        stream_type = meta.get('stream_type') or 'DirectPut'
        arn = self.connection.create_delivery_stream(
            stream_name=name, s3_configuration=s3_configuration,
            stream_type=stream_type,
            kinesis_stream_source=kinesis_configuration)
        _LOG.info(f'Created firehose stream {arn}')
        return self.describe_stream(name=name, meta=meta, arn=arn)

    def describe_stream(self, name, meta, arn, response: dict = None):
        if not response:
            response = self.connection.describe_delivery_stream(name)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def delete_streams(self, args):
        return self.create_pool(self._delete_stream, args)

    @unpack_kwargs
    def _delete_stream(self, arn, config):
        name = config['resource_name']
        response = self.connection.delete_delivery_stream(name)
        return response



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.cloud_front_connection')


@apply_methods_decorator(retry())
class CloudFrontConnection(object):
    """ CloudFront Connection class"""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.client = client('cloudfront', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Cloudfront connection.')

    def get_distribution_config(self, distribution_id):
        """ Crescribes cloud front distribution configuration.

        :type distribution_id: str
        :return: distribution
        """
        return self.client.get_distribution_config(Id=distribution_id)

    def create_distribution(self, distribution_config):
        """ Creates cloud front distribution.

        :type distribution_config: dict
        :return: distribution
        """
        return self.client.create_distribution(
            DistributionConfig=distribution_config)

    def create_invalidation(self, distribution_id, items, quantity,
                            caller_reference):
        inv_params = {
            'Paths':
                {
                    'Quantity': quantity,
                    'Items': items
                },
            'CallerReference': caller_reference
        }
        return self.client.create_invalidation(DistributionId=distribution_id,
                                               InvalidationBatch=inv_params)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import threading
from functools import wraps
from pathlib import PurePath
import click

from syndicate.commons.log_helper import get_logger

_LOG = get_logger('syndicate.core.decorators')

lock = threading.Lock()


def check_deploy_name_for_duplicates(func):
    """
    Checks whether output file with specified name already exists.
    Everywhere this decorator is used the following
    :param func:
    :return:
    """

    @wraps(func)
    def real_wrapper(*args, **kwargs):
        from syndicate.core import CONN
        from syndicate.core import CONFIG
        deploy_name = kwargs.get('deploy_name')
        bundle_name = kwargs.get('bundle_name')
        replace_output = kwargs.get('replace_output')
        if deploy_name and bundle_name and not replace_output:
            output_file_name = f'{bundle_name}/outputs/{deploy_name}.json'

            key_compound = PurePath(CONFIG.deploy_target_bucket_key_compound,
                                    output_file_name).as_posix()
            exists = CONN.s3().is_file_exists(
                CONFIG.deploy_target_bucket,
                key=key_compound)
            if exists:
                _LOG.warn(f'Output file already exists with name '
                          f'{output_file_name}. If it should be replaced with '
                          f'new one, use --replace_output flag.')
                click.echo(f'Output file already exists with name '
                           f'{output_file_name}. If it should be replaced '
                           f'with new one, use --replace_output flag.')
                return
        return func(*args, **kwargs)

    return real_wrapper


def check_deploy_bucket_exists(func):
    @wraps(func)
    def real_wrapper(*args, **kwargs):
        from syndicate.core import CONN
        from syndicate.core import CONFIG
        if not CONN.s3().is_bucket_exists(CONFIG.deploy_target_bucket):
            click.echo(
                f'Cannot execute command: deploy target bucket does not exist.'
                f' Please create it before executing commands that require '
                f'files to be uploaded to the bucket.')
            return
        return func(*args, **kwargs)

    return real_wrapper


def threading_lock(func):
    """ Synchronize access to a function with a threading lock
    to avoid race condition."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        with lock:
            return func(*args, **kwargs)
    return wrapper
